"""Sweeper Module."""

import logging
import typing as t

from zhinst.core import SweeperModule as ZISweeperModule

from zhinst.toolkit.driver.modules.base_module import BaseModule
from zhinst.toolkit.nodetree import Node

if t.TYPE_CHECKING:
    from zhinst.toolkit.session import Session

logger = logging.getLogger(__name__)


class SweeperModule(BaseModule):
    """Implements a base Sweeper Module for Lock-In instruments.

    The Sweeper Module allows for simple and efficient parameter sweeps while
    acquiring data streams from multiple different signal sources. The module
    supports well defined sweeps of various parameters as well as application
    specific measurement presets. For more information on how to use the Sweeper
    Module, have a look at the LabOne Programming Manual.

    For a complete documentation see the LabOne user manual
    https://docs.zhinst.com/labone_programming_manual/sweeper_module.html

    Args:
        sweeper_module: Instance of the core Sweeper Module.
        session: Session to the Data Server.
    """

    def __init__(self, sweeper_module: ZISweeperModule, session: "Session"):
        super().__init__(sweeper_module, session)
        self.root.update_nodes(
            {
                "/gridnode": {
                    "GetParser": self._get_node,
                    "SetParser": self._set_node,
                }
            },
            raise_for_invalid_node=False,
        )

    def execute(self) -> None:
        """Start the sweeper.

        Subscribe or unsubscribe is not possible until the sweep is finished.
        """
        self._raw_module.execute()

    def read(self) -> t.Dict[Node, t.List]:
        """Read the acquired data from the module.

        The data is split into bursts.

        Returns:
            Result of the burst grouped by the signals.
        """
        raw_result = self._raw_module.read(flat=True)
        results = {}
        for node, node_results_raw in raw_result.items():
            results[self._get_node(node)] = node_results_raw
        return results



from unittest.mock import patch

import pytest


@pytest.fixture()
def scope(shfqa):
    yield shfqa.scopes[0]


def test_run(mock_connection, scope):
    # already enabled
    mock_connection.return_value.getInt.return_value = 1
    scope.run()
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/scopes/0/enable", True
    )
    # never disabled
    mock_connection.return_value.getInt.return_value = 0
    with pytest.raises(TimeoutError) as e_info:
        scope.run(timeout=0.5)


def test_stop(mock_connection, scope):
    # already disabled
    mock_connection.return_value.getInt.return_value = 0
    scope.stop()
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/scopes/0/enable", False
    )
    # never disabled
    mock_connection.return_value.getInt.return_value = 1
    with pytest.raises(TimeoutError) as e_info:
        scope.stop(timeout=0.5)


def test_wait_done(mock_connection, scope):
    # already disabled
    mock_connection.return_value.getInt.return_value = 0
    scope.wait_done()
    # never disabled
    mock_connection.return_value.getInt.return_value = 1
    with pytest.raises(TimeoutError) as e_info:
        scope.wait_done(timeout=0.5)


def test_configure(mock_connection, scope):
    with patch("zhinst.toolkit.driver.nodes.shfqa_scope.utils", autospec=True) as utils:
        scope.configure(
            input_select=scope.available_inputs[0],
            num_samples=10,
            trigger_input=scope.available_trigger_inputs[0],
        )
        utils.configure_scope.assert_called_with(
            mock_connection.return_value,
            "DEV1234",
            input_select=scope.available_inputs[0],
            num_samples=10,
            trigger_input=scope.available_trigger_inputs[0],
            num_segments=1,
            num_averages=1,
            trigger_delay=0,
        )
        scope.configure(
            input_select=scope.available_inputs[0],
            num_samples=10,
            trigger_input=scope.available_trigger_inputs[0],
            num_segments=0,
            num_averages=10,
            trigger_delay=33,
        )
        utils.configure_scope.assert_called_with(
            mock_connection.return_value,
            "DEV1234",
            input_select=scope.available_inputs[0],
            num_samples=10,
            trigger_input=scope.available_trigger_inputs[0],
            num_segments=0,
            num_averages=10,
            trigger_delay=33,
        )


def test_read(mock_connection, scope):
    with patch("zhinst.toolkit.driver.nodes.shfqa_scope.utils", autospec=True) as utils:
        utils.get_scope_data.return_value = (
            ["data1", "data2"],
            ["range1", "range2"],
            ["time1", "time2"],
        )
        recorded_data, recorded_data_range, scope_time = scope.read()
        utils.get_scope_data.assert_called_with(
            mock_connection.return_value, "DEV1234", timeout=10
        )
        assert recorded_data == ["data1", "data2"]
        assert recorded_data_range == ["range1", "range2"]
        assert scope_time == ["time1", "time2"]

        # timout argument is forwarded
        scope.read(timeout=1)
        utils.get_scope_data.assert_called_with(
            mock_connection.return_value, "DEV1234", timeout=1
        )



"""Helper script for developers.

The script creates a symbolic link between existing toolkit and
Python environment directories.

By running the script, zhinst-toolkit is added to the ``zhinst`` package namespace
for the given Python environment.

Windows: Requires administration privileges.

zhinst-toolkit must be installed before running the script.

Running the script:

    >>> python scripts/zhinst_toolkit_symlink.py
"""
import os
import sysconfig
from pathlib import Path


SRC_DIR = Path(Path(__file__).parent.parent / Path("src/zhinst/toolkit")).resolve()
DEST_DIR = Path(sysconfig.get_path("purelib")) / "zhinst/toolkit"


def create_symlink(src: Path, dest: Path) -> None:
    """Create symbolic link between working dir and python environment.

    Create symbolic link between existing toolkit and
    Python environment directories.

    Args:
        src: Toolkit directory (e.g. a/b/src/zhinst/toolkit).
        dest: Python environment directory
    """
    try:
        # Windows: Requires administrator when running tests without symlink
        # Otherwise Symlink must be made manually to add toolkit to Python site-packages
        os.symlink(src, dest)
        print(f"Symlink created. Source: {src}, Destination: {dest}")
    except FileExistsError:
        print(f"Destination: {dest} already exists.")


if __name__ == "__main__":
    create_symlink(SRC_DIR, DEST_DIR)



"""zhinst-toolkit spectroscopy node adaptions."""
import logging

import numpy as np
import zhinst.utils.shfqa as utils

from zhinst.toolkit.interface import AveragingMode
from zhinst.toolkit.nodetree import Node, NodeTree

logger = logging.getLogger(__name__)


class Spectroscopy(Node):
    """Spectroscopy node.

    Implements basic functionality of the spectroscopy, e.g allowing the user to
    read the result logger data.

    Args:
        root: Root of the nodetree
        tree: Tree (node path as tuple) of the current node
        serial: Serial of the device.
        index: Index of the corresponding awg channel
    """

    def __init__(
        self,
        root: NodeTree,
        tree: tuple,
        serial: str,
        index: int,
    ):
        super().__init__(root, tree)
        self._daq_server = root.connection
        self._serial = serial
        self._index = index

    def configure_result_logger(
        self,
        *,
        result_length: int,
        num_averages: int = 1,
        averaging_mode: AveragingMode = AveragingMode.CYCLIC,
    ) -> None:
        """Configures the result logger for spectroscopy mode.

        Args:
            result_length: Number of results to be returned by the result logger
            num_averages: Number of averages, will be rounded to 2^n.
            averaging_mode: Averaging order of the result.
        """
        utils.configure_result_logger_for_spectroscopy(
            self._daq_server,
            self._serial,
            self._index,
            result_length=result_length,
            num_averages=num_averages,
            averaging_mode=int(averaging_mode),
        )

    def run(self) -> None:
        """Resets and enables the spectroscopy result logger."""
        utils.enable_result_logger(
            self._daq_server,
            self._serial,
            self._index,
            mode="spectroscopy",
        )

    def stop(self, *, timeout: float = 10, sleep_time: float = 0.05) -> None:
        """Stop the result logger.

        Args:
            timeout: The maximum waiting time in seconds for the
                Spectroscopy (default: 10).
            sleep_time: Time in seconds to wait between
                requesting Spectroscopy state
        Raises:
            TimeoutError: If the result logger could not been stopped within the
                given time.

        """
        self.result.enable(False)
        try:
            self.result.enable.wait_for_state_change(
                0, timeout=timeout, sleep_time=sleep_time
            )
        except TimeoutError as error:
            raise TimeoutError(
                f"{repr(self)}: The result logger could not been stopped "
                f"within the specified timeout ({timeout}s)."
            ) from error

    def wait_done(self, *, timeout: float = 10, sleep_time: float = 0.05) -> None:
        """Wait until spectroscopy is finished.

        Args:
            timeout (float): The maximum waiting time in seconds for the
                Spectroscopy (default: 10).
            sleep_time (float): Time in seconds to wait between
                requesting Spectroscopy state

        Raises:
            TimeoutError: if the spectroscopy recording is not completed within the
                given time.

        """
        try:
            self.result.enable.wait_for_state_change(
                0, timeout=timeout, sleep_time=sleep_time
            )
        except TimeoutError as error:
            raise TimeoutError(
                f"{repr(self)}: The spectroscopy did not finish "
                f"within the specified timeout ({timeout}s)."
            ) from error

    def read(
        self,
        *,
        timeout: float = 10,
    ) -> np.array:
        """Waits until the logger finished recording and returns the measured data.

        Args:
            timeout: Maximum time to wait for data in seconds (default = 10s)

        Returns:
            An array containing the result logger data.

        """
        return utils.get_result_logger_data(
            self._daq_server,
            self._serial,
            self._index,
            mode="spectroscopy",
            timeout=timeout,
        )



"""zhinst-toolkit AWG node adaptions."""
import json
import logging
import typing as t

from zhinst.core import compile_seqc
from zhinst.toolkit.driver.nodes.command_table_node import CommandTableNode
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.nodetree.helper import (
    lazy_property,
    create_or_append_set_transaction,
)
from zhinst.toolkit.waveform import Waveforms
from zhinst.toolkit.sequence import Sequence

logger = logging.getLogger(__name__)


class AWG(Node):
    """AWG node.

    This class implements the basic functionality for the device specific
    arbitrary waveform generator.
    Besides the upload/compilation of sequences it offers the upload of
    waveforms and command tables.

    Args:
        root: Root of the nodetree
        tree: Tree (node path as tuple) of the current node
        session: Underlying session.
        serial: Serial of the device.
        index: Index of the corresponding awg channel
        device_type: Device type
    """

    def __init__(
        self,
        root: NodeTree,
        tree: tuple,
        serial: str,
        index: int,
        device_type: str,
        device_options: str,
    ):
        Node.__init__(self, root, tree)
        self._daq_server = root.connection
        self._serial = serial
        self._index = index
        self._device_type = device_type
        self._device_options = device_options

    def enable_sequencer(self, *, single: bool) -> None:
        """Starts the sequencer of a specific channel.

        Waits until the sequencer is enabled.

        Args:
            single: Flag if the sequencer should be disabled after finishing
            execution.
        """
        self.single(single)
        self.enable(1, deep=True)
        self.enable.wait_for_state_change(1)

    def wait_done(self, *, timeout: float = 10, sleep_time: float = 0.005) -> None:
        """Wait until the AWG is finished.

        Args:
            timeout: The maximum waiting time in seconds for the generator
                (default: 10).
            sleep_time: Time in seconds to wait between requesting generator
                state

        Raises:
            RuntimeError: If continuous mode is enabled
            TimeoutError: If the sequencer program did not finish within
                the specified timeout time
        """
        if not self.single():
            raise RuntimeError(
                f"{repr(self)}: The generator is running in continuous mode, "
                "it will never be finished."
            )
        try:
            self.enable.wait_for_state_change(0, timeout=timeout, sleep_time=sleep_time)
        except TimeoutError as error:
            raise TimeoutError(
                f"{repr(self)}: The execution of the sequencer program did not finish "
                f"within the specified timeout ({timeout}s)."
            ) from error

    def compile_sequencer_program(
        self,
        sequencer_program: t.Union[str, Sequence],
        **kwargs: t.Union[str, int],
    ) -> t.Tuple[bytes, t.Dict[str, t.Any]]:
        """Compiles a sequencer program for the specific device.

        Args:
            sequencer_program: The sequencer program to compile.

        Keyword Args:
            samplerate: Target sample rate of the sequencer. Only allowed/
                necessary for HDAWG devices. Must correspond to the samplerate
                used by the device (device.system.clocks.sampleclock.freq()).
                If not specified the function will get the value itself from
                the device. It is recommended passing the samplerate if more
                than one sequencer code is uploaded in a row to speed up the
                execution time.
            wavepath: path to directory with waveforms. Defaults to path used
                by LabOne UI or AWG Module.
            waveforms: waveform CSV files separated by ';'
            output: name of embedded ELF filename.

        Returns:
            elf: Binary ELF data for sequencer.
            extra: Extra dictionary with compiler output.

        Examples:
            >>> elf, compile_info = device.awgs[0].compile_sequencer_program(seqc)
            >>> device.awgs[0].elf.data(elf)
            >>> device.awgs[0].ready.wait_for_state_change(1)
            >>> device.awgs[0].enable(True)

        Raises:
            RuntimeError: `sequencer_program` is empty.
            RuntimeError: If the compilation failed.

        .. versionadded:: 0.4.0
        """
        if "SHFQC" in self._device_type:
            kwargs["sequencer"] = "sg" if "sgchannels" in self._tree else "qa"
        elif "HDAWG" in self._device_type and "samplerate" not in kwargs:
            kwargs["samplerate"] = self.root.system.clocks.sampleclock.freq()

        return compile_seqc(
            str(sequencer_program),
            self._device_type,
            self._device_options,
            **kwargs,
        )

    def load_sequencer_program(
        self,
        sequencer_program: t.Union[str, Sequence],
        **kwargs: t.Union[str, int],
    ) -> t.Dict[str, t.Any]:
        """Compiles the given sequencer program on the AWG Core.

        Warning:
            After uploading the sequencer program one needs to wait before for
            the awg core to become ready before it can be enabled.
            The awg core indicates the ready state through its `ready` node.
            (device.awgs[0].ready() == True)

        Args:
            sequencer_program: Sequencer program to be uploaded.

        Keyword Args:
            samplerate: Target sample rate of the sequencer. Only allowed/
                necessary for HDAWG devices. Must correspond to the samplerate
                used by the device (device.system.clocks.sampleclock.freq()).
                If not specified the function will get the value itself from
                the device. It is recommended passing the samplerate if more
                than one sequencer code is uploaded in a row to speed up the
                execution time.
            wavepath: path to directory with waveforms. Defaults to path used
                by LabOne UI or AWG Module.
            waveforms: waveform CSV files separated by ';'
            output: name of embedded ELF filename.

        Examples:
            >>> compile_info = device.awgs[0].load_sequencer_program(seqc)
            >>> device.awgs[0].ready.wait_for_state_change(1)
            >>> device.awgs[0].enable(True)

        Raises:
            RuntimeError: `sequencer_program` is empty.
            RuntimeError: If the upload or compilation failed.

        .. versionadded:: 0.3.4

            `sequencer_program` does not accept empty strings

        .. versionadded:: 0.4.0

            Use offline compiler instead of AWG module to compile the sequencer
            program. This speeds of the compilation and also enables parallel
            compilation/upload.
        """
        elf, compiler_info = self.compile_sequencer_program(sequencer_program, **kwargs)
        self.elf.data(elf)
        return compiler_info

    def write_to_waveform_memory(
        self, waveforms: Waveforms, indexes: list = None, validate: bool = True
    ) -> None:
        """Writes waveforms to the waveform memory.

        The waveforms must already be assigned in the sequencer program.

        Args:
            waveforms: Waveforms that should be uploaded.
            indexes: Specify a list of indexes that should be uploaded. If
                nothing is specified all available indexes in waveforms will
                be uploaded. (default = None)
            validate: Enable sanity check preformed by toolkit, based on the
                waveform descriptors on the device. Can be disabled for e.g.
                speed optimizations. Does not affect the checks happen in LabOne
                and or the firmware. (default = True)

        Raises:
            IndexError: The index of a waveform exceeds the one on the device
                and `validate` is True.
            RuntimeError: One of the waveforms index points to a
                filler(placeholder) and `validate` is True.
        """
        waveform_info = None
        num_waveforms = None
        if validate:
            waveform_info = json.loads(self.waveform.descriptors()).get("waveforms", [])
            num_waveforms = len(waveform_info)
        with create_or_append_set_transaction(self._root):
            for waveform_index in waveforms.keys():
                if indexes and waveform_index not in indexes:
                    continue
                if num_waveforms is not None and waveform_index >= num_waveforms:
                    raise IndexError(
                        f"There are {num_waveforms} waveforms defined on the device "
                        "but the passed waveforms specified one with index "
                        f"{waveform_index}."
                    )
                if (
                    waveform_info
                    and "__filler" in waveform_info[waveform_index]["name"]
                ):
                    raise RuntimeError(
                        f"The waveform at index {waveform_index} is only "
                        "a filler and can not be overwritten"
                    )
                self.root.transaction.add(
                    self.waveform.waves[waveform_index],
                    waveforms.get_raw_vector(
                        waveform_index,
                        target_length=int(waveform_info[waveform_index]["length"])
                        if waveform_info
                        else None,
                    ),
                )

    def read_from_waveform_memory(self, indexes: t.List[int] = None) -> Waveforms:
        """Read waveforms to the waveform memory.

        Args:
            indexes: List of waveform indexes to read from the device. If not
                specified all assigned waveforms will be downloaded.

        Returns:
            Waveform object with the downloaded waveforms.
        """
        nodes = [self.waveform.descriptors.node_info.path]
        if indexes is not None:
            for index in indexes:
                nodes.append(self.waveform.node_info.path + f"/waves/{index}")
        else:
            nodes.append(self.waveform.waves["*"].node_info.path)
        nodes_str = ",".join(nodes)
        waveforms_raw = self._daq_server.get(nodes_str, settingsonly=False, flat=True)
        waveform_info = json.loads(
            waveforms_raw.pop(self.waveform.descriptors.node_info.path)[0]["vector"]
        ).get("waveforms", [])
        waveforms = Waveforms()
        for node, waveform in waveforms_raw.items():
            slot = int(node[-1])
            if "__filler" not in waveform_info[slot]["name"]:
                waveforms.assign_native_awg_waveform(
                    slot,
                    waveform[0]["vector"],
                    channels=int(waveform_info[slot].get("channels", 1)),
                    markers_present=bool(
                        int(waveform_info[slot].get("marker_bits")[0])
                    ),
                )
        return waveforms

    @lazy_property
    def commandtable(self) -> t.Optional[CommandTableNode]:
        """Command table module."""
        if self["commandtable"].is_valid():
            return CommandTableNode(
                self._root, self._tree + ("commandtable",), self._device_type
            )
        return None



"""SHFQA Instrument Driver."""

import logging

from zhinst.toolkit.driver.devices import SHFQA, SHFSG

logger = logging.getLogger(__name__)


class SHFQC(SHFQA, SHFSG):
    """High-level driver for the Zurich Instruments SHFQC."""



import json
from unittest.mock import patch

import jsonref
import jsonschema
import pytest

from zhinst.toolkit.command_table import CommandTable, ParentNode, _derefence_json
from zhinst.toolkit.exceptions import ValidationError


def test_ct_clear(command_table_schema):
    ct2 = CommandTable(command_table_schema)
    ct = CommandTable(command_table_schema)
    ct.table[0].amplitude00.value = 1
    ct.header.userString = "Foo bar"
    ct.clear()
    assert ct.as_dict() == ct2.as_dict()


def test_initialize_command_table_dict_json(command_table_schema):
    obj = CommandTable(command_table_schema)
    assert isinstance(obj._ct_schema, dict)


def test_initialize_command_table_str_json(command_table_schema):
    obj = CommandTable(json.dumps(command_table_schema))
    assert isinstance(obj._ct_schema, dict)


def test_parent_node_repr():
    foo = ParentNode({}, ("table", "0", "bar", "foo"))
    assert str(foo) == "/table/0/bar/foo"


def test_command_table_header_version_22_02(command_table_schema_22_02):
    ct = CommandTable(command_table_schema_22_02)
    assert (
        ct.header.version
        == command_table_schema_22_02["definitions"]["header"]["properties"]["version"][
            "enum"
        ][0]
    )


def test_command_table_header_version_22_08(command_table_schema_22_08):
    ct = CommandTable(command_table_schema_22_08)
    assert ct.header.version == command_table_schema_22_08["version"]


def test_properties_parent_entry(command_table_schema):
    ct = CommandTable(command_table_schema)
    test_props = dir(ct.table[0])
    json_props = list(command_table_schema["definitions"]["entry"]["properties"].keys())
    assert test_props.sort() == json_props.sort()


def test_parent_entry_get_attr_underscore(command_table):
    assert command_table.table[0]._foobar is None


def test_parent_entry_get_property_attr(command_table):
    assert command_table.table[0].amplitude00.value is None


def test_list_entry_len(command_table):
    command_table.table[0]
    command_table.table[1]
    command_table.table[5]
    command_table.table[1]
    assert len(command_table.table) == 3


def test_parent_entry_contains(command_table):
    assert "value" in command_table.table[0].amplitude00
    assert "waveform" in command_table.table[0]
    assert "version" in command_table.header
    assert "canno" not in command_table.header


def test_header_parent_entry(command_table_schema):
    ct = CommandTable(command_table_schema)
    test_props = dir(ct.header)
    json_props = list(
        command_table_schema["definitions"]["header"]["properties"].keys()
    )
    assert test_props.sort() == json_props.sort()


def test_parent_entry_properties(command_table_schema):
    ct = CommandTable(command_table_schema)
    for property_ in command_table_schema["definitions"]["entry"]["properties"].keys():
        if property_ != "index":
            assert hasattr(ct.table[0], property_) is True


def test_property_not_existing(command_table):
    with pytest.raises(AttributeError):
        getattr(command_table.table[0], "this_property_cannot_exists")


def test_parent_entry_header(command_table_schema):
    ct = CommandTable(command_table_schema)
    r = jsonref.loads(json.dumps(command_table_schema))
    for property_ in r["definitions"]["header"]["properties"].keys():
        assert (
            ct.header.info(property_)
            == r["definitions"]["header"]["properties"][property_]
        )


def test_assert_validate_called_table_index(command_table):
    with patch("zhinst.toolkit.command_table.jsonschema.validate") as mocked_method:
        command_table.table[44]
        mocked_method.assert_called_once_with(
            instance=44,
            schema={
                "type": "integer",
                "minimum": 0,
                "maximum": 1023,
                "exclusiveMinimum": False,
                "exclusiveMaximum": False,
            },
            cls=jsonschema.validators.Draft4Validator,
        )


@pytest.mark.parametrize("input_, output", [(1, 1), (40, 40), ("foo", "foo"), (-2, -2)])
def test_assert_validate_called_parent_entry_attribute_set(
    input_, output, command_table
):
    obj = command_table.table[0]
    with patch("zhinst.toolkit.command_table.jsonschema.validate") as mocked_method:
        obj.amplitude00.value = input_
        mocked_method.assert_called_once_with(
            instance=output,
            schema={
                "description": "Amplitude scaling factor of the given AWG channel",
                "type": "number",
                "minimum": -1.0,
                "maximum": 1.0,
                "exclusiveMinimum": False,
                "exclusiveMaximum": False,
            },
            cls=jsonschema.validators.Draft4Validator,
        )


def test_parent_entry_table_index(command_table):
    with pytest.raises(AttributeError):
        command_table.table[0].index = 5321


def test_parent_entry_table_waveform(command_table):
    command_table.table[0].waveform.index = 2000
    assert command_table.table[0].waveform.index == 2000


def test_build_json_no_items(command_table):
    assert command_table.as_dict() == {
        "$schema": "https://json-schema.org/draft-04/schema#",
        "header": {"version": "1.1.0"},
        "table": [],
    }


def test_build_json_items(command_table_schema):
    command_table = CommandTable(command_table_schema)
    assert command_table.as_dict() == {
        "$schema": "https://json-schema.org/draft-04/schema#",
        "header": {"version": "1.1.0"},
        "table": [],
    }
    command_table.table[0].amplitude00.value = 1
    command_table.table[0].waveform.index = 0

    command_table.table[2].amplitude00.value = 0.5
    command_table.table[2].waveform.index = 3
    command_table.table[2].waveform.awgChannel0 = ["sigout1"]
    assert command_table.as_dict() == {
        "$schema": command_table_schema["$schema"],
        "header": {"version": "1.1.0"},
        "table": [
            {"amplitude00": {"value": 1}, "index": 0, "waveform": {"index": 0}},
            {
                "amplitude00": {"value": 0.5},
                "index": 2,
                "waveform": {"awgChannel0": ["sigout1"], "index": 3},
            },
        ],
    }


def test_table_index_value_error(command_table_schema):
    command_table = CommandTable(command_table_schema)
    idx_min = command_table_schema["definitions"]["tableindex"]["minimum"]
    command_table_schema["definitions"]["tableindex"]["maximum"]

    with pytest.raises(ValidationError):
        command_table.table[idx_min - 5]
    command_table.table[idx_min + 1]


def test_load_completed_command_table(command_table_schema, command_table_completed):
    ct = CommandTable(command_table_schema)
    ct.update(command_table_completed)
    assert command_table_completed == ct.as_dict()

    ct.update(json.dumps(command_table_completed))
    assert command_table_completed == ct.as_dict()


def test_modify_completed_command_table(command_table_schema, command_table_completed):
    ct = CommandTable(command_table_schema)
    ct.update(command_table_completed)
    assert ct.table[0].amplitude00.as_dict() == {
        "increment": False,
        "value": 0,
    }
    ct.table[0].amplitude00.value = 1
    ct.table[0].amplitude00.increment = True
    assert ct.table[0].amplitude00.as_dict() == {
        "increment": True,
        "value": 1,
    }


def test_command_table_schema_version_mismatch(
    command_table_schema, command_table_completed
):
    command_table_completed["header"]["version"] = "2.0"
    ct = CommandTable(command_table_schema)
    with pytest.raises(ValidationError):
        ct.update(command_table_completed)


def test_parent_delete_attribute(command_table):
    command_table.table[0].amplitude00.value = 1
    command_table.table[0].waveform.index = 10
    assert command_table.table[0].as_dict() == {
        "amplitude00": {"value": 1},
        "waveform": {"index": 10},
    }
    command_table.table[0].amplitude00 = None
    assert command_table.table[0].as_dict() == {"waveform": {"index": 10}}


def test_parent_delete_attribute_with_none(command_table):
    command_table.table[0].amplitude00.value = 1
    assert command_table.table[0].as_dict() == {
        "amplitude00": {"value": 1},
    }
    command_table.table[0].amplitude00 = None
    assert command_table.table[0].as_dict() == {}


def test_parent_property_delete_attribute_with_none(command_table):
    command_table.table[0].amplitude00.value = 1
    assert command_table.table[0].as_dict() == {
        "amplitude00": {"value": 1},
    }
    command_table.table[0].amplitude00.value = None
    assert command_table.table[0].as_dict() == {}


def test_json_non_existing_childs_ignored(command_table):
    command_table.table[0].amplitude00
    command_table.table[0].amplitude00.value
    command_table.table[0]
    command_table
    command_table.header
    command_table.header.userString
    assert command_table.as_dict() == {
        "$schema": "https://json-schema.org/draft-04/schema#",
        "header": {"version": "1.1.0"},
        "table": [],
    }


def test_parent_clear(command_table):
    command_table.table[0].amplitude00.value = 1
    command_table.table[0].waveform.index = 10
    command_table.table[0].clear()
    assert command_table.table[0].as_dict() == {}


def test_list_entry_delete(command_table):
    command_table.table[2].amplitude00.value = 1
    del command_table.table[2]
    assert command_table.as_dict()["table"] == []


def test_test_entry_list_get_range(command_table):
    assert command_table.table.range == (0, 1024)


def test_maximum_items_in_table(command_table):
    idx = command_table.table.range[1] + 1
    with pytest.raises(ValidationError):
        command_table.table[idx]


@pytest.mark.parametrize("value", ["bar", 1, 1.2, []])
def test_derefence_json(value):
    with pytest.raises(ValueError):
        _derefence_json(value)


def test_minimum_items_in_table(command_table):
    idx = command_table.table.range[0] - 1
    with pytest.raises(ValidationError):
        command_table.table[idx]


def test_table_property_validation_error(command_table):
    with pytest.raises(ValidationError):
        command_table.table[0].amplitude00.value = 99999



import numpy as np
import pytest

from zhinst.toolkit import Waveforms
from zhinst.toolkit.driver.devices.uhfqa import QAS, UHFQA, Integration
from zhinst.toolkit.driver.nodes.awg import AWG
from zhinst.toolkit.nodetree import Node


@pytest.fixture()
def uhfqa(data_dir, mock_connection, session):

    json_path = data_dir / "nodedoc_dev1234_uhfqa.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.listNodesJSON.return_value = nodes_json
    mock_connection.return_value.getString.return_value = ""

    yield UHFQA("DEV1234", "UHFQA", session)


def test_repr(uhfqa):
    assert repr(uhfqa) == "UHFQA(UHFQA,DEV1234)"


def test_qas(uhfqa):
    assert len(uhfqa.qas) == 1
    assert isinstance(uhfqa.qas[0], QAS)
    # Wildcards nodes will be converted into normal Nodes
    assert not isinstance(uhfqa.qas["*"], QAS)
    assert isinstance(uhfqa.qas["*"], Node)


def test_qas_integration(uhfqa):
    assert isinstance(uhfqa.qas[0].integration, Integration)
    assert isinstance(uhfqa.qas["*"].integration, Node)


def test_enable_qccs_mode(mock_connection, uhfqa):
    uhfqa.enable_qccs_mode()
    mock_connection.return_value.set.assert_called_with(
        [
            ("/dev1234/system/extclk", "external"),
            ("/dev1234/dios/0/extclk", "internal"),
            ("/dev1234/dios/0/mode", "qa_result_qccs"),
            ("/dev1234/dios/0/drive", 3),
            ("/dev1234/awgs/0/dio/strobe/slope", "off"),
            ("/dev1234/awgs/0/dio/valid/index", 16),
            ("/dev1234/awgs/0/dio/valid/polarity", "high"),
        ]
    )
    mock_connection.reset_mock()
    with uhfqa.set_transaction():
        uhfqa.enable_qccs_mode()
    mock_connection.return_value.set.assert_called_with(
        [
            ("/dev1234/system/extclk", "external"),
            ("/dev1234/dios/0/extclk", "internal"),
            ("/dev1234/dios/0/mode", "qa_result_qccs"),
            ("/dev1234/dios/0/drive", 3),
            ("/dev1234/awgs/0/dio/strobe/slope", "off"),
            ("/dev1234/awgs/0/dio/valid/index", 16),
            ("/dev1234/awgs/0/dio/valid/polarity", "high"),
        ]
    )


def test_adjusted_delay(uhfqa, mock_connection):

    default_delay = 0
    bypass_deskew = 0

    def get_int_side_effect(node):
        if "/dev1234/qas/0/delay" in node:
            return default_delay
        if "/dev1234/qas/0/bypass/deskew" in node:
            return bypass_deskew
        raise RuntimeError("Node not found")

    mock_connection.return_value.getInt.side_effect = get_int_side_effect

    # get value
    default_delay = 200
    assert uhfqa.qas[0].adjusted_delay() == 0
    default_delay = 300
    assert uhfqa.qas[0].adjusted_delay() == 100
    bypass_deskew = 1
    default_delay = 200
    assert uhfqa.qas[0].adjusted_delay() == 16
    default_delay = 300
    assert uhfqa.qas[0].adjusted_delay() == 116

    # set value
    bypass_deskew = 0
    assert uhfqa.qas[0].adjusted_delay(400) == 400
    mock_connection.return_value.set.assert_called_with(f"/dev1234/qas/0/delay", 600)
    bypass_deskew = 1
    assert uhfqa.qas[0].adjusted_delay(400) == 400
    mock_connection.return_value.set.assert_called_with(f"/dev1234/qas/0/delay", 584)

    with pytest.raises(ValueError):
        uhfqa.qas[0].adjusted_delay(-185)

    with pytest.raises(ValueError):
        uhfqa.qas[0].adjusted_delay(1025 - 185)


def test_qas_crosstalk_matrix(uhfqa, mock_connection):

    matrix = np.random.rand(10, 10)

    def get_double_side_effect(node):
        if "/dev1234/qas/0/crosstalk/rows/" in node:
            return matrix[int(node[-8]), int(node[-1])]
        raise RuntimeError("Node not found")

    mock_connection.return_value.getDouble.side_effect = get_double_side_effect

    # get
    result = uhfqa.qas[0].crosstalk_matrix()
    assert np.allclose(result, matrix)

    # set
    matrix2 = np.random.rand(10, 10)
    uhfqa.qas[0].crosstalk_matrix(matrix2)
    for row in range(10):
        for col in range(10):
            mock_connection.return_value.set.assert_any_call(
                f"/dev1234/qas/0/crosstalk/rows/{row}/cols/{col}", matrix2[row][col]
            )

    matrix2 = np.random.rand(11, 10)
    with pytest.raises(ValueError):
        uhfqa.qas[0].crosstalk_matrix(matrix2)


def test_awg(data_dir, mock_connection, uhfqa):
    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    mock_connection.return_value.getString.return_value = "AWG,FOOBAR"
    assert len(uhfqa.awgs) == 1
    assert isinstance(uhfqa.awgs[0], AWG)
    # Wildcards nodes will be converted into normal Nodes
    assert not isinstance(uhfqa.awgs["*"], AWG)
    assert isinstance(uhfqa.awgs["*"], Node)


class TestWriteIntegrationWeights:
    def test_write_integration_weights_waveforms(self, mock_connection, uhfqa):
        wave1 = 1.0 * np.ones(8)
        wave2 = -1.0 * np.ones(8)
        waveforms = Waveforms()
        waveforms.assign_waveform(0, wave1, wave2)
        uhfqa.qas[0].integration.write_integration_weights(waveforms)
        call_args = mock_connection.return_value.set.call_args
        assert "/dev1234/qas/0/integration/weights/0/real" == call_args[0][0][0][0]
        np.testing.assert_array_equal(wave1, call_args[0][0][0][1])
        assert "/dev1234/qas/0/integration/weights/0/imag" == call_args[0][0][1][0]
        np.testing.assert_array_equal(wave2, call_args[0][0][1][1])

    def test_write_integration_weights_waveforms_idx1(self, mock_connection, uhfqa):
        wave1 = 1.0 * np.ones(8)
        waveforms = Waveforms()
        waveforms.assign_waveform(1, wave1)
        uhfqa.qas[0].integration.write_integration_weights(waveforms)
        call_args = mock_connection.return_value.set.call_args
        assert "/dev1234/qas/0/integration/weights/1/real" == call_args[0][0][0][0]
        np.testing.assert_array_equal(wave1, call_args[0][0][0][1])

    def test_write_integration_weights_dict_complex(self, mock_connection, uhfqa):
        wave_complex = 1.0 * np.ones(8, dtype=np.complex_)
        uhfqa.qas[0].integration.write_integration_weights({0: wave_complex})
        call_args = mock_connection.return_value.set.call_args
        assert "/dev1234/qas/0/integration/weights/0/real" == call_args[0][0][0][0]
        np.testing.assert_array_equal(wave_complex.real, call_args[0][0][0][1])
        assert "/dev1234/qas/0/integration/weights/0/imag" == call_args[0][0][1][0]
        np.testing.assert_array_equal(wave_complex.imag, call_args[0][0][1][1])

    def test_write_integration_weights_dict_real(self, mock_connection, uhfqa):
        wave = 1.0 * np.ones(8)
        uhfqa.qas[0].integration.write_integration_weights({0: wave})
        call_args = mock_connection.return_value.set.call_args
        assert "/dev1234/qas/0/integration/weights/0/real" == call_args[0][0][0][0]
        np.testing.assert_array_equal(wave.real, call_args[0][0][0][1])
        assert "/dev1234/qas/0/integration/weights/0/imag" == call_args[0][0][1][0]
        np.testing.assert_array_equal(np.zeros(8), call_args[0][0][1][1])



"""Module for managing a session to a Data Server through zhinst.core."""
import json
import typing as t
from collections.abc import MutableMapping
from enum import IntFlag
from pathlib import Path
from contextlib import contextmanager

import zhinst.toolkit.driver.devices as tk_devices
import zhinst.toolkit.driver.modules as tk_modules
from zhinst import core
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.nodetree.helper import lazy_property
from zhinst.toolkit.nodetree.nodetree import Transaction


class Devices(MutableMapping):
    """Mapping class for the connected devices.

    Maps the connected devices from data server to lazy device objects.
    On every access the connected devices are read from the data server. This
    ensures that even if devices get connected/disconnected through another
    session the list will be up to date.

    Args:
        session: An active session to the data server.
    """

    def __init__(self, session: "Session"):
        self._session = session
        self._devices: t.Dict[str, tk_devices.DeviceType] = {}
        self._device_classes = tk_devices.DEVICE_CLASS_BY_MODEL

    def __getitem__(self, key) -> tk_devices.DeviceType:
        key = key.lower()
        if key in self.connected():
            if key not in self._devices:
                self._devices[key] = self._create_device(key)
                # start a transaction if the session has a ongoing one
                if self._session.multi_transaction.in_progress():
                    self._devices[key].root.transaction.start(
                        self._session.multi_transaction.add
                    )
            return self._devices[key]
        self._devices.pop(key, None)
        raise KeyError(key)

    def __setitem__(self, *_):
        raise LookupError(
            "Illegal operation. Can not add a device manually. Devices must be "
            "connected through the session (session.connect_device)."
        )

    def __delitem__(self, key):
        self._devices.pop(key, None)

    def __iter__(self):
        return iter(self.connected())

    def __len__(self):
        return len(self.connected())

    def _create_device(self, serial: str) -> tk_devices.DeviceType:
        """Creates a new device object.

        Maps the device type to the correct instrument class (The default is
        the ``BaseInstrument`` which is a generic instrument class that supports
        all devices).

        Warning:
            The device must be connected to the data server.

        Args:
            serial: Device serial

        Returns:
            Newly created instrument object

        Raises:
            RuntimeError: If the device is not connected to the data server
        """
        dev_type = self._session.daq_server.getString(f"/{serial}/features/devtype")
        return self._device_classes.get(dev_type, tk_devices.BaseInstrument)(
            serial, dev_type, self._session
        )

    def connected(self) -> t.List[str]:
        """Get a list of devices connected to the data server.

        Returns:
            List of all connected devices.
        """
        return (
            self._session.daq_server.getString("/zi/devices/connected")
            .lower()
            .split(",")
        )

    def visible(self) -> t.List[str]:
        """Get a list of devices visible to the data server.

        Returns:
            List of all connected devices.
        """
        return (
            self._session.daq_server.getString("/zi/devices/visible").lower().split(",")
        )

    def created_devices(self) -> t.ValuesView[tk_devices.DeviceType]:
        """View on all created device.

        The list contains all toolkit device objects that have been created for
        the underlying session.

        Warning: This is not equal to the devices connected to the data server!
            Use the iterator of the `Devices` class directly to get all devices
            connected to the data server.
        """
        return self._devices.values()


class HF2Devices(Devices):
    """Mapping class for the connected HF2 devices.

    Maps the connected devices from data server to lazy device objects.
    It derives from the general ``Devices`` class and adds the special handling
    for the HF2 data server. Since the HF2 Data Server is based on the API Level
    1 it as a much more restricted API. This means it is not possible to get
    the connected or visible devices from the data server. This class must
    track the connected devices itself and use discovery to mimic the
    behavior of the new data server used for the other devices.
    """

    def _create_device(self, serial: str) -> tk_devices.BaseInstrument:
        """Creates a new device object.

        Maps the device type to the correct instrument class (The default is
        the ``BaseInstrument`` which is a generic instrument class that supports
        all devices).

        Warning:
            The device must already be connected to the data server

        Args:
            serial: Device serial

        Returns:
            Newly created instrument object

        Raises:
            RuntimeError: If the device is not connected to the data server
        """
        try:
            return super()._create_device(serial)
        except RuntimeError as error:
            if "ZIAPINotFoundException" in error.args[0]:
                discovery = core.ziDiscovery()
                discovery.find(serial)
                dev_type = discovery.get(serial)["devicetype"]
                raise RuntimeError(
                    "Can only connect HF2 devices to an HF2 data "
                    f"server. {serial} identifies itself as a {dev_type}."
                ) from error
            raise

    def connected(self) -> t.List[str]:
        """Get a list of devices connected to the data server.

        Returns:
            List of all connected devices.
        """
        return list(self._devices.keys())

    def visible(self) -> t.List[str]:
        """Get a list of devices visible to the data server.

        Returns:
            List of all connected devices.
        """
        return core.ziDiscovery().findAll()

    def add_hf2_device(self, serial: str) -> None:
        """Add a new HF2 device.

        Since the HF2 data server is not able to report its connected devices
        toolkit manually needs to update the list of known connected devices.

        Args:
            serial: Serial of the HF2 device

        Raises:
            RuntimeError: If the device was already added in that session.
        """
        if serial in self._devices:
            raise RuntimeError(f"Can only create one instance of {serial}.")
        self._devices[serial] = self._create_device(serial)


class ModuleHandler:
    """Modules of LabOne.

    Handler for all additional so called modules by LabOne. A LabOne module is
    bound to a user session but creates a independent session to the Data Server.
    This has the advantage that they do not interfere with the user session. It
    also means that creating a session causes additional resources allocation,
    both at the client and the data server. New modules should therefore only be
    instantiated with care.

    Toolkit holds a lazy generated instance of all modules. This ensures that
    not more than one modules of each type gets created by accident and that the
    access to the modules is optimized.

    Of course there are many use cases where more than one module of a single
    type is required. This class therefore also exposes a ``create`` function for
    each LabOne module. These functions create a unmanaged instance of that
    module (unmanaged means toolkit does not hold an instance of that module).

    Args:
        session: Active user session
        server_host: Host address of the session
        server_port: Port of the session
    """

    def __init__(self, session: "Session"):
        self._session = session

    def __repr__(self):
        return str(
            "LabOneModules("
            f"{self._session.daq_server.host}:{self._session.daq_server.port})"
        )

    def create_awg_module(self) -> tk_modules.BaseModule:
        """Create an instance of the AwgModule.

        The resulting Module will have the nodetree accessible. The underlying
        zhinst.core Module can be accessed through the `raw_module`
        property.

        The new instance establishes a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `awg`.

        Returns:
            Created module
        """
        return tk_modules.BaseModule(
            self._session.daq_server.awgModule(), self._session
        )

    def create_daq_module(self) -> tk_modules.DAQModule:
        """Create an instance of the DataAcquisitionModule.

        The resulting Module will have the nodetree accessible. The underlying
        zhinst.core Module can be accessed through the `raw_module`
        property.

        The new instance establishes a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `daq`.

        Returns:
            Created module
        """
        return tk_modules.DAQModule(
            self._session.daq_server.dataAcquisitionModule(), self._session
        )

    def create_device_settings_module(self) -> tk_modules.BaseModule:
        """Create an instance of the DeviceSettingsModule.

        The resulting Module will have the nodetree accessible. The underlying
        zhinst.core Module can be accessed through the `raw_module`
        property.

        The new instance establishes a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `device_settings_module`.

        Returns:
            Created module
        """
        return tk_modules.BaseModule(
            self._session.daq_server.deviceSettings(), self._session
        )

    def create_impedance_module(self) -> tk_modules.BaseModule:
        """Create an instance of the ImpedanceModule.

        The resulting Module will have the nodetree accessible. The underlying
        zhinst.core Module can be accessed through the `raw_module`
        property.

        The new instance establishes a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `impedance_module`.

        Returns:
            Created module
        """
        return tk_modules.BaseModule(
            self._session.daq_server.impedanceModule(), self._session
        )

    def create_mds_module(self) -> tk_modules.BaseModule:
        """Create an instance of the MultiDeviceSyncModule.

        The resulting Module will have the nodetree accessible. The underlying
        zhinst.core Module can be accessed through the `raw_module`
        property.

        The new instance establishes a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `mds_module`.

        Returns:
            Created module
        """
        return tk_modules.BaseModule(
            self._session.daq_server.multiDeviceSyncModule(), self._session
        )

    def create_pid_advisor_module(self) -> tk_modules.BaseModule:
        """Create an instance of the PidAdvisorModule.

        The resulting Module will have the nodetree accessible. The underlying
        zhinst.core Module can be accessed through the `raw_module`
        property.

        The new instance establishes a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `pid_advisor_module`.

        Returns:
            Created module
        """
        return tk_modules.BaseModule(
            self._session.daq_server.pidAdvisor(), self._session
        )

    def create_precompensation_advisor_module(
        self,
    ) -> tk_modules.BaseModule:
        """Create an instance of the PrecompensationAdvisorModule.

        In contrast to core.ziDAQServer.precompensationAdvisor() a nodetree property
        is added.

        The new instance establishes a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `precompensation_advisor_module`.

        Returns:
            Created module
        """
        return tk_modules.BaseModule(
            self._session.daq_server.precompensationAdvisor(), self._session
        )

    def create_qa_module(self) -> tk_modules.BaseModule:
        """Create an instance of the QuantumAnalyzerModule.

        The resulting Module will have the nodetree accessible. The underlying
        zhinst.core Module can be accessed through the `raw_module`
        property.

        The new instance establishes a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `qa_module`.

        Returns:
            Created module
        """
        return tk_modules.BaseModule(
            self._session.daq_server.quantumAnalyzerModule(), self._session
        )

    def create_scope_module(self) -> tk_modules.BaseModule:
        """Create an instance of the ScopeModule.

        The resulting Module will have the nodetree accessible. The underlying
        zhinst.core Module can be accessed through the `raw_module`
        property.

        The new instance establishes a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `awg_module`.

        Returns:
            Created module
        """
        return tk_modules.BaseModule(
            self._session.daq_server.scopeModule(), self._session
        )

    def create_sweeper_module(self) -> tk_modules.SweeperModule:
        """Create an instance of the SweeperModule.

        The resulting Module will have the nodetree accessible. The underlying
        zhinst.core Module can be accessed through the `raw_module`
        property.

        The new instance establishes a new session to the DataServer.
        New instances should therefor be created carefully since they consume
        resources.

        The new module is not managed by toolkit. A managed instance is provided
        by the property `sweeper_module`.

        Returns:
            Created module
        """
        return tk_modules.SweeperModule(self._session.daq_server.sweep(), self._session)

    def create_shfqa_sweeper(self) -> tk_modules.SHFQASweeper:
        """Create an instance of the SHFQASweeper.

        For now the general sweeper module does not support the SHFQA. However a
        python based implementation called ``SHFSweeper`` does already provide
        this functionality. The ``SHFSweeper`` is part of the ``zhinst`` module
        and can be found in the utils.

        Toolkit wraps around the ``SHFSweeper`` and exposes a interface that is
        similar to the LabOne modules, meaning the parameters are exposed in a
        node tree like structure.

        In addition a new session is created. This has the benefit that the
        sweeper implementation does not interfere with the the commands and
        setups from the user.

        Returns:
            Created object
        """
        return tk_modules.SHFQASweeper(self._session)

    @lazy_property
    def awg(self) -> tk_modules.BaseModule:
        """Managed instance of the awg module.

        Managed means that only one instance is created
        and is held inside the connection Manager. This makes it easier to access
        the modules from within toolkit, since creating a module requires
        resources. (``use create_awg_module`` to create an unmanaged instance)
        """
        return self.create_awg_module()

    @lazy_property
    def daq(self) -> tk_modules.DAQModule:
        """Managed instance of the daq module.

        Managed means that only one instance is created
        and is held inside the connection Manager. This makes it easier to access
        the modules from within toolkit, since creating a module requires
        resources. (``use create_daq_module`` to create an unmanaged instance)
        """
        return self.create_daq_module()

    @lazy_property
    def device_settings(self) -> tk_modules.BaseModule:
        """Managed instance of the device settings module.

        Managed means that only one instance is created
        and is held inside the connection Manager. This makes it easier to access
        the modules from within toolkit, since creating a module requires
        resources. (``use create_device_settings_module`` to create an
        unmanaged instance)
        """
        return self.create_device_settings_module()

    @lazy_property
    def impedance(self) -> tk_modules.BaseModule:
        """Managed instance of the impedance module.

        Managed means that only one instance is created
        and is held inside the connection Manager. This makes it easier to access
        the modules from within toolkit, since creating a module requires
        resources. (``use create_awg_module`` to create an unmanaged instance)
        """
        return self.create_impedance_module()

    @lazy_property
    def mds(self) -> tk_modules.BaseModule:
        """Managed instance of the multi device sync module.

        Managed means that only one instance is created
        and is held inside the connection Manager. This makes it easier to access
        the modules from within toolkit, since creating a module requires
        resources. (``use create_mds_module`` to create an unmanaged instance)
        """
        return self.create_mds_module()

    @lazy_property
    def pid_advisor(self) -> tk_modules.BaseModule:
        """Managed instance of the pid advisor module.

        Managed means that only one instance is created
        and is held inside the connection Manager. This makes it easier to access
        the modules from within toolkit, since creating a module requires
        resources. (``use create_pid_advisor_module`` to create an unmanaged
        instance)
        """
        return self.create_pid_advisor_module()

    @lazy_property
    def precompensation_advisor(self) -> tk_modules.BaseModule:
        """Managed instance of the precompensation advisor module.

        Managed means that only one instance is created
        and is held inside the connection Manager. This makes it easier to access
        the modules from within toolkit, since creating a module requires
        resources. (``use create_precompensation_advisor_module`` to create an
        unmanaged instance)
        """
        return self.create_precompensation_advisor_module()

    @lazy_property
    def qa(self) -> tk_modules.BaseModule:
        """Managed instance of the quantum analyzer module.

        Managed means that only one instance is created
        and is held inside the connection Manager. This makes it easier to access
        the modules from within toolkit, since creating a module requires
        resources. (``use create_qa_module`` to create an unmanaged instance)
        """
        return self.create_qa_module()

    @lazy_property
    def scope(self) -> tk_modules.BaseModule:
        """Managed instance of the scope module.

        Managed means that only one instance is created
        and is held inside the connection Manager. This makes it easier to access
        the modules from within toolkit, since creating a module requires
        resources. (``use create_scope_module`` to create an unmanaged
        instance)
        """
        return self.create_scope_module()

    @lazy_property
    def sweeper(self) -> tk_modules.SweeperModule:
        """Managed instance of the sweeper module.

        Managed means that only one instance is created
        and is held inside the connection Manager. This makes it easier to access
        the modules from within toolkit, since creating a module requires
        resources. (``use create_sweeper_module`` to create an unmanaged instance)
        """
        return self.create_sweeper_module()

    @lazy_property
    def shfqa_sweeper(self) -> tk_modules.SHFQASweeper:
        """Managed instance of the shfqa sweeper implementation.

        Managed means that only one instance is created
        and is held inside the connection Manager. This makes it easier to access
        the modules from within toolkit, since creating a module requires
        resources. (``use create_shfqa_sweeper`` to create an unmanaged
        instance)
        """
        return self.create_shfqa_sweeper()


class PollFlags(IntFlag):
    """Flags for polling Command.

    DETECT_AND_THROW(12):
        Detect data loss holes and throw EOFError exception

    DETECT(8):
        Detect data loss holes

    FILL(1):
        Fill holes

    DEFAULT(0):
        No Flags

    Can be combined with bitwise operations

    >>> PollFlags.FILL | PollFlags.DETECT
        <PollFlags.DETECT|FILL: 9>
    """

    DETECT_AND_THROW = 12
    DETECT = 8
    FILL = 1
    DEFAULT = 0


class Session(Node):
    """Session to a data server.

    Zurich Instruments devices use a server-based connectivity methodology.
    Server-based means that all communication between the user and the
    instrument takes place via a computer program called a server, the data
    sever. The data sever recognizes available instruments and manages all
    communication between the instrument and the host computer on one side, and
    communication to all the connected clients on the other side. (For more
    information on the architecture please refer to the user manual
    http://docs.zhinst.com/labone_programming_manual/introduction.html)

    The entry point into for any connection is therefore a client session to a
    existing data sever. This class represents a single client session to a
    data server. The session enables the user to connect to one or multiple
    instruments (also creates the dedicated objects for each device), access
    the LabOne modules and poll data. In short it is the only object the user
    need to create by himself.

    Info:
        Except for the HF2 a single session can be used to connect to all
        devices from Zurich Instruments. Since the HF2 is historically based on
        another data server called the hf2 data server it is not possible to
        connect HF2 devices a "normal" data server and also not possible to
        connect devices apart from HF2 to the hf2 data server.

    Args:
        server_host: Host address of the data server (e.g. localhost)
        server_port: Port number of the data server. If not specified the session
            uses the default port 8004 (8005 for HF2 if specified).
            (default = None)
        hf2: Flag if the session should be established with an HF2 data sever or
            the "normal" one for all other devices. If not specified the session
            will detect the type of the data server based on the port.
            (default = None)
        connection: Existing DAQ server object. If specified the session will
            not create a new session to the data server but reuse the passed
            one. (default = None)
    """

    def __init__(
        self,
        server_host: str,
        server_port: t.Optional[int] = None,
        *,
        hf2: t.Optional[bool] = None,
        connection: t.Optional[core.ziDAQServer] = None,
    ):
        self._is_hf2_server = bool(hf2)
        if connection is not None:
            self._is_hf2_server = "HF2" in connection.getString("/zi/about/dataserver")
            if hf2 and not self._is_hf2_server:
                raise RuntimeError(
                    "hf2 flag was set but the passed "
                    "DAQServer instance is not a HF2 data server."
                )
            if hf2 is False and self._is_hf2_server:
                raise RuntimeError(
                    "hf2 flag was set but the passed "
                    "DAQServer instance is a HF2 data server."
                )
            self._daq_server = connection
        else:
            server_port = server_port if server_port else 8004
            if self._is_hf2_server and server_port == 8004:
                server_port = 8005
            try:
                self._daq_server = core.ziDAQServer(
                    server_host,
                    server_port,
                    1 if self._is_hf2_server else 6,
                )
            except RuntimeError as error:
                if "Unsupported API level" not in error.args[0]:
                    raise
                if hf2 is None:
                    self._is_hf2_server = True
                    self._daq_server = core.ziDAQServer(
                        server_host,
                        server_port,
                        1,
                    )
                elif not hf2:
                    raise RuntimeError(
                        "hf2 Flag was reset but the specified "
                        f"server at {server_host}:{server_port} is a "
                        "HF2 data server."
                    ) from error

        if self._is_hf2_server and "HF2" not in self._daq_server.getString(
            "/zi/about/dataserver"
        ):
            raise RuntimeError(
                "hf2 Flag was set but the specified "
                f"server at {server_host}:{server_port} is not a "
                "HF2 data server."
            )
        self._devices = HF2Devices(self) if self._is_hf2_server else Devices(self)
        self._modules = ModuleHandler(self)

        hf2_node_doc = Path(__file__).parent / "resources/nodedoc_hf2_data_server.json"
        nodetree = NodeTree(
            self._daq_server,
            prefix_hide="zi",
            list_nodes=["/zi/*"],
            preloaded_json=json.loads(hf2_node_doc.open("r").read())
            if self._is_hf2_server
            else None,
        )
        super().__init__(nodetree, tuple())
        self._multi_transaction = Transaction(self.root)

    def __repr__(self):
        return str(
            f"{'HF2' if self._is_hf2_server else ''}DataServerSession("
            f"{self._daq_server.host}:{self._daq_server.port})"
        )

    @classmethod
    def from_existing_connection(cls, connection: core.ziDAQServer) -> "Session":
        """Initialize Session from an existing connection.

        Args:
            connection: Existing connection.

        .. versionadded:: 0.4.0
        """
        is_hf2_server = "HF2" in connection.getString("/zi/about/dataserver")
        return cls(
            server_host=connection.host,
            server_port=connection.port,
            hf2=is_hf2_server,
            connection=connection,
        )

    def connect_device(
        self, serial: str, *, interface: t.Optional[str] = None
    ) -> tk_devices.DeviceType:
        """Establish a connection to a device.

        Info:
            It is allowed to call this function for an already connected device.
            In that case the function simply returns the device object of the
            device.

        Args:
            serial: Serial number of the device, e.g. *'dev12000'*.
                The serial number can be found on the back panel of the
                instrument.
            interface: Device interface (e.g. = "1GbE"). If not specified
                the default interface from the discover is used.

        Returns:
            Device object

        Raises:
            KeyError: Device is not found.
            RuntimeError: Connection failed.
        """
        serial = serial.lower()
        if serial not in self._devices:
            if not interface:
                if self._is_hf2_server:
                    interface = "USB"
                else:
                    # Take interface from the discovery
                    interface = json.loads(self.daq_server.getString("/zi/devices"))[
                        serial.upper()
                    ]["INTERFACE"]
            try:
                self._daq_server.connectDevice(serial, interface)
            except RuntimeError as error:
                if (
                    interface == "PCIe"
                    and "Device does not support the specified interface"
                    in error.args[0]
                ):
                    self._daq_server.connectDevice(serial, "1GbE")
                else:
                    raise
            if isinstance(self._devices, HF2Devices):
                self._devices.add_hf2_device(serial)
        return self._devices[serial]

    def disconnect_device(self, serial: str) -> None:
        """Disconnect a device.

        Warning:
            This function will return immediately. The disconnection of the
            device may not yet finished.

        Args:
            serial: Serial number of the device, e.g. *'dev12000'*.
                The serial number can be found on the back panel of the instrument.
        """
        self._devices.pop(serial, None)
        self.daq_server.disconnectDevice(serial)

    def sync(self) -> None:
        """Synchronize all connected devices.

        Synchronization in this case means creating a defined state.

        The following steps are performed:
            * Ensures that all set commands have been flushed to the device
            * Ensures that get and poll commands only return data which was
              recorded after the sync command. (ALL poll buffers are cleared!)
            * Blocks until all devices have cleared their busy flag.

        Warning:
            The sync is performed for all devices connected to the DAQ server

        Warning:
            This command is a blocking command that can take a substantial
            amount of time.

        Raises:
            RuntimeError: ZIAPIServerException: Timeout during sync of device
        """
        self.daq_server.sync()

    def poll(
        self,
        recording_time: float = 0.1,
        *,
        timeout: float = 0.5,
        flags: PollFlags = PollFlags.DEFAULT,
    ) -> t.Dict[Node, t.Dict[str, t.Any]]:
        """Polls all subscribed data from the data server.

        Poll the value changes in all subscribed nodes since either subscribing
        or the last poll (assuming no buffer overflow has occurred on the Data
        Server).

        Args:
            recording_time: Defines the duration of the poll in seconds. (Note that not
                only the newly recorded values are polled but all values since
                either subscribing or the last poll). Needs to be larger than
                zero. (default = 0.1)
            timeout: Adds an additional timeout in seconds on top of
                `recording_time`. Only relevant when communicating in a slow
                network. In this case it may be set to a value larger than the
                expected round-trip time in the network. (default = 0.5)
            flags: Flags for the polling (see :class `PollFlags`:)

        Returns:
            Polled data in a dictionary. The key is a `Node` object and the
            value is a dictionary with the raw data from the device
        """
        data_raw = self.daq_server.poll(
            recording_time, int(timeout * 1000), flags=flags.value, flat=True
        )
        return {self.raw_path_to_node(node): data for node, data in data_raw.items()}

    def raw_path_to_node(
        self, raw_path: str, *, module: tk_modules.ModuleType = None
    ) -> Node:
        """Converts a raw node path string into a Node object.

        The device that this strings belongs to must be connected to the Data
        Server. Optionally a module can be specified to which the node belongs to.
        (The module is only an additional search path, meaning even if a module
        is specified the node can belong to a connected device.)

        Args:
            raw_path: Raw node path (e.g. /dev1234/relative/path/to/node).

        Returns:
            Corresponding toolkit node object.

        Raises:
            RuntimeError: If the node does not belong to the optional module or
                to a connected device.
        """
        if not raw_path.startswith("/"):
            raise RuntimeError(
                f"{raw_path} does not seem to be an absolute path. "
                "(it must start with a leading slash)"
            )
        if module is not None:
            node = module.root.raw_path_to_node(raw_path)
            if node.raw_tree[0] in module.root:
                return node
        try:
            serial = raw_path.split("/")[1]
            if serial == "zi":
                return self.root.raw_path_to_node(raw_path)
            return self.devices[serial].root.raw_path_to_node(raw_path)
        except KeyError as error:
            raise RuntimeError(
                f"Node belongs to a device({raw_path.split('/')[1]}) not connected to "
                "the Data Server."
            ) from error

    @contextmanager
    def set_transaction(self) -> t.Generator[None, None, None]:
        """Context manager for a transactional set.

        Can be used as a context in a with statement and bundles all node set
        commands into a single transaction. This reduces the network overhead
        and often increases the speed.

        In comparison to the device level transaction manager this manager
        affects all devices that are connected to the Session and bundles all
        set commands into a single transaction.

        Within the with block a set commands to a node will be buffered
        and bundled into a single command at the end automatically.
        (All other operations, e.g. getting the value of a node, will not be
        affected)

        Warning:
            The set is always performed as deep set if called on device nodes.

        Examples:
            >>> with session.set_transaction():
                    device1.test[0].a(1)
                    device2.test[0].a(2)

        .. versionadded:: 0.4.0
        """
        self._multi_transaction.start()
        for device in self.devices.created_devices():
            device.root.transaction.start(self._multi_transaction.add)
        self.root.transaction.start(self._multi_transaction.add)
        try:
            yield
            self._daq_server.set(self._multi_transaction.result())
        finally:
            for device in self.devices.created_devices():
                device.root.transaction.stop()
            self.root.transaction.stop()
            self._multi_transaction.stop()

    @property
    def multi_transaction(self) -> Transaction:
        """Flag if a session wide transaction is in progress.

        .. versionadded:: 0.4.0
        """
        return self._multi_transaction

    @property
    def devices(self) -> Devices:
        """Mapping for the connected devices."""
        return self._devices

    @property
    def modules(self) -> ModuleHandler:
        """Modules of LabOne."""
        return self._modules

    @property
    def is_hf2_server(self) -> bool:
        """Flag if the data server is a HF2 Data Server."""
        return self._is_hf2_server

    @property
    def daq_server(self) -> core.ziDAQServer:
        """Managed instance of the core.ziDAQServer."""
        return self._daq_server

    @property
    def server_host(self) -> str:
        """Server host."""
        return self._daq_server.host

    @property
    def server_port(self) -> int:
        """Server port."""
        return self._daq_server.port



import pytest

from zhinst.toolkit.driver.devices.uhfli import AWG, UHFLI
from zhinst.toolkit.nodetree import Node


@pytest.fixture()
def uhfli(data_dir, mock_connection, session):

    json_path = data_dir / "nodedoc_dev1234_uhfli.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.listNodesJSON.return_value = nodes_json
    mock_connection.return_value.getString.return_value = ""

    yield UHFLI("DEV1234", "UHFLI", session)


def test_repr(uhfli):
    assert repr(uhfli) == "UHFLI(UHFLI,DEV1234)"


def test_awg(data_dir, mock_connection, uhfli):
    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    mock_connection.return_value.getString.return_value = "AWG,FOOBAR"
    assert len(uhfli.awgs) == 1
    assert isinstance(uhfli.awgs[0], AWG)
    # Wildcards nodes will be converted into normal Nodes
    assert not isinstance(uhfli.awgs["*"], AWG)
    assert isinstance(uhfli.awgs["*"], Node)

    # test command table
    assert not uhfli.awgs[0].commandtable


def test_awg_option_not_found(data_dir, mock_connection, uhfli):
    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    mock_connection.return_value.getString.return_value = "FOOBAR"
    assert len(uhfli.awgs) == 1
    assert isinstance(uhfli.awgs[0], Node)



"""zhinst-toolkit readout node adaptions."""
import logging
import typing as t

import numpy as np
import zhinst.utils.shfqa as utils

from zhinst.toolkit.interface import AveragingMode
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.waveform import Waveforms

logger = logging.getLogger(__name__)


class Readout(Node):
    """Readout node.

    Implements basic functionality of the readout, e.g allowing the user to
    write the integration weight.

    Args:
        root: Root of the nodetree
        tree: Tree (node path as tuple) of the current node
        serial: Serial of the device.
        index: Index of the corresponding awg channel
        max_qubits_per_channel: Max qubits per channel
    """

    def __init__(
        self,
        root: NodeTree,
        tree: tuple,
        serial: str,
        index: int,
        max_qubits_per_channel: int,
    ):
        super().__init__(root, tree)
        self._daq_server = root.connection
        self._serial = serial
        self._index = index
        self._max_qubits_per_channel = max_qubits_per_channel

    def configure_result_logger(
        self,
        *,
        result_source: str,
        result_length: int,
        num_averages: int = 1,
        averaging_mode: AveragingMode = AveragingMode.CYCLIC,
    ) -> None:
        """Configures the result logger for readout mode.

        Args:
            result_source: String-based tag to select the result source in readout
                mode, e.g. "result_of_integration" or "result_of_discrimination".
            result_length: Number of results to be returned by the result logger
            num_averages: Number of averages, will be rounded to 2^n
            averaging_mode: Select the averaging order of the result, with
                0 = cyclic and 1 = sequential.
        """
        utils.configure_result_logger_for_readout(
            self._daq_server,
            self._serial,
            self._index,
            result_source=result_source,
            result_length=result_length,
            num_averages=num_averages,
            averaging_mode=int(averaging_mode),
        )

    def run(self) -> None:
        """Reset and enable the result logger."""
        utils.enable_result_logger(
            self._daq_server,
            self._serial,
            self._index,
            mode="readout",
        )

    def stop(self, *, timeout: float = 10, sleep_time: float = 0.05) -> None:
        """Stop the result logger.

        Args:
            timeout: The maximum waiting time in seconds for the Readout
                (default: 10).
            sleep_time: Sleep interval in seconds. (default = 0.05)

        Raises:
            TimeoutError: The result logger could not been stopped within the
                given time.
        """
        self.result.enable(False)
        try:
            self.result.enable.wait_for_state_change(
                0, timeout=timeout, sleep_time=sleep_time
            )
        except TimeoutError as error:
            raise TimeoutError(
                f"{repr(self)}: The result logger could not been stopped "
                f"within the specified timeout ({timeout}s)."
            ) from error

    def wait_done(self, *, timeout: float = 10, sleep_time: float = 0.05) -> None:
        """Wait until the readout is finished.

        Args:
            timeout: The maximum waiting time in seconds for the Readout
                (default: 10).
            sleep_time: Sleep interval in seconds. (default = 0.05)

        Raises:
            TimeoutError: if the readout recording is not completed within the
                given time.
        """
        try:
            self.result.enable.wait_for_state_change(
                0, timeout=timeout, sleep_time=sleep_time
            )
        except TimeoutError as error:
            raise TimeoutError(
                f"{repr(self)}: The readout did not finish "
                f"within the specified timeout ({timeout}s)."
            ) from error

    def read(
        self,
        *,
        timeout: float = 10,
    ) -> np.array:
        """Waits until the logger finished recording and returns the measured data.

        Args:
            timeout: Maximum time to wait for data in seconds (default = 10s)

        Returns:
            Result logger data.
        """
        return utils.get_result_logger_data(
            self._daq_server, self._serial, self._index, mode="readout", timeout=timeout
        )

    def write_integration_weights(
        self,
        weights: t.Union[Waveforms, dict],
        *,
        integration_delay: float = 0.0,
        clear_existing: bool = True,
    ) -> None:
        """Configures the weighted integration.

        Args:
            weights: Dictionary containing the complex weight vectors, where
                keys correspond to the indices of the integration units to be
                configured.
            integration_delay: Delay in seconds before starting the readout.
                (default = 0.0)
            clear_existing: Flag whether to clear the waveform memory before
                the present upload. (default = True)
        """
        if (
            len(weights.keys()) > 0
            and max(weights.keys()) >= self._max_qubits_per_channel
        ):
            raise RuntimeError(
                f"The device only has {self._max_qubits_per_channel} qubits per channel"
                f", but {max(weights.keys())} where specified"
            )
        waveform_dict = {}
        if isinstance(weights, Waveforms):
            for slot in weights.keys():
                waveform_dict[slot] = weights.get_raw_vector(slot, complex_output=True)
        else:
            waveform_dict = weights

        utils.configure_weighted_integration(
            self._daq_server,
            self._serial,
            self._index,
            weights=waveform_dict,
            integration_delay=integration_delay,
            clear_existing=clear_existing,
        )

    def read_integration_weights(self, slots: t.List[int] = None) -> Waveforms:
        """Read integration weights from the waveform memory.

        Args:
            slots: List of weight slots to read from the device. If not specified
                all available weights will be downloaded.

        Returns:
            Mutable mapping of the downloaded weights.
        """
        nodes = []
        if slots is not None:
            for slot in slots:
                nodes.append(self.integration.weights[slot].wave.node_info.path)
        else:
            nodes.append(self.integration.weights["*"].wave.node_info.path)
        nodes_str = ",".join(nodes)
        weights_raw = self._daq_server.get(nodes_str, settingsonly=False, flat=True)
        weights = Waveforms()
        for slot, weight in enumerate(weights_raw.values()):
            weights[slot] = weight[0]["vector"]
        return weights



import numpy as np
import pytest

from zhinst.toolkit.waveform import Waveforms, Wave, OutputType


def test_dict_behavior():
    waveform = Waveforms()
    wave1 = 1.0 * np.ones(1008)
    wave1_short = 1.0 * np.ones(500)
    wave2 = -1.0 * np.ones(1008)
    wave3 = -0.5 * np.ones(1008)
    marker = 0.0 * np.ones(1008)

    with pytest.raises(TypeError) as e_info:
        waveform[0] = 1
    with pytest.raises(RuntimeError) as e_info:
        waveform[0] = (wave1, wave2, wave3, marker)

    # "standart" waveform
    waveform[0] = (wave1, wave2)
    assert all(waveform[0][0] == wave1)
    assert all(waveform[0][1] == wave2)
    assert waveform[0][2] == None
    assert len(waveform.get_raw_vector(0)) == 1008 * 2

    # replace wave
    waveform[0] = (wave1, wave3)
    assert all(waveform[0][0] == wave1)
    assert all(waveform[0][1] == wave3)
    assert waveform[0][2] == None

    # replace wave
    waveform.assign_waveform(0, wave1, wave2)
    assert all(waveform[0][0] == wave1)
    assert all(waveform[0][1] == wave2)
    assert waveform[0][2] == None

    # delete wave
    assert 0 in waveform.keys()
    del waveform[0]
    assert 0 not in waveform.keys()

    # iter
    waveform[0] = (wave1, wave3)
    waveform[2] = (wave1, wave3)
    waveform[10] = (wave1, wave3)
    assert len(waveform) == 3
    num_elements = 0
    for _, element in waveform.items():
        assert all(element[0] == wave1)
        assert all(element[1] == wave3)
        num_elements += 1
    assert num_elements == len(waveform)

    # "standart" waveform with marker
    waveform[1] = (wave1, wave2, marker)
    assert all(waveform[1][0] == wave1)
    assert all(waveform[1][1] == wave2)
    assert all(waveform[1][2] == marker)
    assert len(waveform.get_raw_vector(1)) == 1008 * 3

    # unequal length
    with pytest.raises(RuntimeError) as e_info:
        waveform[10] = (wave1_short, wave2)
    with pytest.raises(RuntimeError) as e_info:
        waveform[10] = (wave1, wave2, wave1_short)

    # invalid inputs
    with pytest.raises(RuntimeError) as e_info:
        waveform[10] = (0, 0, 0)

    # complex values
    waveform[10] = np.ones(1008, dtype=np.complex128)
    waveform[10] = (np.ones(1008, dtype=np.complex128), marker)
    waveform[10] = (np.ones(1008, dtype=np.complex128), None, marker)
    with pytest.raises(RuntimeError) as e_info:
        waveform[10] = (np.ones(1008, dtype=np.complex128), wave2, None)


def test_get_raw_vector():
    waveform = Waveforms()
    waveform[0] = (np.ones(1008), np.ones(1008))
    waveform[1] = (np.ones(1008), np.ones(1008), np.ones(1008))
    waveform[2] = np.ones(1008, dtype=np.complex128)
    waveform[3] = (np.ones(1008, dtype=np.complex128), np.ones(1008))

    # non complex
    result0 = waveform.get_raw_vector(0)
    assert len(result0) == 2016
    result1 = waveform.get_raw_vector(1)
    assert len(result1) == 3024
    assert result0[0] == result1[0]
    assert result0[1] == result1[1]
    assert result1[2] == 1
    result2 = waveform.get_raw_vector(2)
    assert len(result2) == 2016
    assert result0[0] == result2[0]
    assert result2[1] == 0
    assert result0[2] == result2[2]
    result3 = waveform.get_raw_vector(3)
    assert len(result3) == 3024
    assert result0[0] == result3[0]
    assert result3[1] == 0
    assert result3[2] == 1

    # complex
    result0 = waveform.get_raw_vector(0, complex_output=True)
    assert len(result0) == 1008
    assert result0.dtype == np.complex128
    with pytest.warns(RuntimeWarning):
        result1 = waveform.get_raw_vector(1, complex_output=True)
    assert len(result1) == 1008
    assert all(result0 == result1)
    result2 = waveform.get_raw_vector(2, complex_output=True)
    assert all(result2 == np.ones(1008, dtype=np.complex128))
    with pytest.warns(RuntimeWarning):
        result3 = waveform.get_raw_vector(3, complex_output=True)
    assert all(result3 == np.ones(1008, dtype=np.complex128))

    # invalid length
    with pytest.raises(ValueError):
        waveform.get_raw_vector(0, target_length=1000)
    with pytest.raises(ValueError):
        waveform.get_raw_vector(0, target_length=2000)


def test_assign():
    waveform = Waveforms()
    wave = np.ones(1008)
    waveform[0] = (wave,)
    assert all(waveform[0][0] == wave)
    waveform[1] = (wave, -wave)
    assert all(waveform[1][0] == wave)
    assert all(waveform[1][1] == -wave)
    waveform[2] = (wave, wave, -wave)
    assert all(waveform[2][0] == wave)
    assert all(waveform[2][1] == wave)
    assert all(waveform[2][2] == -wave)
    waveform[3] = (0.5 * wave, None, wave)
    assert all(waveform[3][0] == 0.5 * wave)
    assert waveform[3][1] == None
    assert all(waveform[3][2] == wave)
    waveform[4] = (np.ones(1008, dtype=np.complex128), wave)
    assert all(waveform[4][0] == np.ones(1008, dtype=np.complex128))
    assert all(waveform[4][1] == wave)
    waveform[6] = (Wave(-wave, "w1", 1), Wave(wave, "w2", 2), wave)
    assert all(waveform[6][0] == -wave)
    assert waveform[6][0].name == "w1"
    assert waveform[6][0].output == 1
    assert all(waveform[6][1] == wave)
    assert waveform[6][1].name == "w2"
    assert waveform[6][1].output == 2
    assert all(waveform[6][2] == wave)
    waveform[7] = (Wave(wave, "w", 1), wave)
    assert all(waveform[7][0] == wave)
    assert waveform[7][0].name == "w"
    assert waveform[7][0].output == 1
    assert all(waveform[2][1] == wave)
    assert waveform[7][1].name == None
    assert waveform[7][1].output == None
    assert waveform[7][2] == None
    waveform[8] = (Wave(wave, "test", 3), wave, -wave)
    assert all(waveform[8][0] == wave)
    assert waveform[8][0].name == "test"
    assert waveform[8][0].output == 3
    assert all(waveform[8][1] == wave)
    assert waveform[8][1].name == None
    assert waveform[8][1].output == None
    assert all(waveform[8][2] == -wave)


def test_sequence_snippet():
    waveform = Waveforms()
    waveform[0] = (Wave(np.ones(1008), "w1", 3), Wave(np.ones(1008), "w2", 2))
    waveform[3] = (Wave(np.ones(252), "w3"), np.ones(252), 1 * np.ones(252))
    waveform[1] = (np.ones(504), np.ones(504), 15 * np.ones(504))
    waveform[5] = (np.ones(252),)

    result = waveform.get_sequence_snippet()
    assert (
        result
        == """\
wave w1 = placeholder(1008, false, false);
wave w2 = placeholder(1008, false, false);
assignWaveIndex(1, 2, w1, 2, w2, 0);
assignWaveIndex(placeholder(504, true, true), placeholder(504, true, true), 1);
wave w3 = placeholder(252, true, false);
assignWaveIndex(w3, placeholder(252, false, false), 3);
assignWaveIndex(placeholder(252, false, false), 5);"""
    )

    waveform = Waveforms()
    waveform.assign_waveform(
        0,
        Wave(np.ones(1008), name="w1", output=OutputType.OUT1 | OutputType.OUT2),
        Wave(np.ones(1008), name="w2", output=OutputType.OUT1 | OutputType.OUT2),
        np.ones(1008),
    )
    waveform.assign_waveform(
        3,
        np.ones(1008),
        np.ones(1008),
        np.ones(1008),
    )
    waveform.assign_waveform(
        5,
        np.ones(1008),
        np.ones(1008),
    )
    waveform.assign_waveform(
        4,
        Wave(np.ones(1008), name="test1", output=OutputType.OUT2),
        Wave(np.ones(1008), name="test2", output=OutputType.OUT1),
    )
    waveform[2] = (np.ones(1008, dtype=np.complex128), None, 15 * np.ones(1008))

    waveform[1] = (
        Wave(
            np.ones(1008, dtype=np.complex128),
            name=["comp1", "comp2"],
            output=[OutputType.OUT2, OutputType.OUT1 | OutputType.OUT2],
        ),
    )

    result = waveform.get_sequence_snippet()
    assert (
        result
        == """\
wave w1 = placeholder(1008, true, false);
wave w2 = placeholder(1008, false, false);
assignWaveIndex(1, 2, w1, 1, 2, w2, 0);
wave comp1 = placeholder(1008, false, false);
wave comp2 = placeholder(1008, false, false);
assignWaveIndex(2, comp1, 1, 2, comp2, 1);
assignWaveIndex(placeholder(1008, true, true), placeholder(1008, true, true), 2);
assignWaveIndex(placeholder(1008, true, false), placeholder(1008, false, false), 3);
wave test1 = placeholder(1008, false, false);
wave test2 = placeholder(1008, false, false);
assignWaveIndex(2, test1, 1, test2, 4);
assignWaveIndex(placeholder(1008, false, false), placeholder(1008, false, false), 5);"""
    )



"""HDAWG Instrument Driver."""
import typing as t

from zhinst.toolkit.driver.devices.base import BaseInstrument
from zhinst.toolkit.driver.nodes.awg import AWG
from zhinst.toolkit.nodetree.helper import (
    create_or_append_set_transaction,
    lazy_property,
)
from zhinst.toolkit.nodetree.node import NodeList


class HDAWG(BaseInstrument):
    """High-level driver for the Zurich Instruments HDAWG."""

    def enable_qccs_mode(self) -> None:
        """Configure the instrument to work with PQSC.

        This method sets the reference clock source to
        connect the instrument to the PQSC.

        Info:
            Use ``factory_reset`` to reset the changes if necessary
        """
        with create_or_append_set_transaction(self._root):
            # Set ZSync clock to be used as reference
            self.system.clocks.referenceclock.source("zsync")
            # Configure DIO
            # Set interface standard to use on the 32-bit DIO to LVCMOS
            self.dios[0].interface(0)
            # Set DIO output values to ZSync input values.
            # Forward the ZSync input values to the AWG sequencer.
            # Forward the DIO input values to the ZSync output.
            self.dios[0].mode("qccs")
            # Drive the two most significant bytes of the DIO port
            self.dios[0].drive(0b1100)
            # Disable DIO triggering on the AWGs,
            # since it's not needed for ZSync messages
            self.awgs["*"].dio.strobe.slope("off")
            self.awgs["*"].dio.valid.polarity("none")

    @lazy_property
    def awgs(self) -> t.Sequence[AWG]:
        """A Sequence of AWG Cores."""
        return NodeList(
            [
                AWG(
                    self.root,
                    self._tree + ("awgs", str(i)),
                    self.serial,
                    i,
                    self.device_type,
                    self.device_options,
                )
                for i in range(len(self["awgs"]))
            ],
            self._root,
            self._tree + ("awgs",),
        )



"""Single lazy node of a :class:`zhinst.toolkit.nodetree.Nodetree`."""

import fnmatch
import json
import numbers
import re
import time
import typing as t
from collections import namedtuple
from collections.abc import Sequence, Mapping
from enum import IntEnum
from functools import lru_cache

from zhinst.toolkit.nodetree.helper import (
    create_or_append_set_transaction,
    lazy_property,
)

if t.TYPE_CHECKING:
    from zhinst.toolkit.nodetree import NodeTree


class NodeEnumMeta:
    """Custom Metaclass for NodeEnum.

    Note: Required to enable pickling of a NodeEnum value.

    It simply servers the purpose to recreate a NodeEnum for a given enum
    value. Since the NodeEnums are created dynamically there is no way recreate
    a NodeEnum value since python can not find the definition. This class
    bypasses this problem by providing the functionality to recreate the
    Enum on the fly.

    Warning: Although the class of the resulting enum object looks and feels
    the same as the original one it is not. Therefore comparing the `type` will
    fail. This is however the only limitation.
    (type(value_old) != type(value_new) but value_old == value_new)

    Args:
        value: Value of the NodeEnum object that should be created.
        class_name: Name of the NodeEnum class.
        names: Mapping of the enum names to their corresponding integer value.
        module: Should be set to the module this class is being created in.
    """

    def __new__(  # noqa: D102
        cls, value: int, class_name: str, names: t.Dict[str, int], module: str
    ):
        new_enum = NodeEnum(class_name, names, module=module)
        return new_enum(value)


class NodeEnum(IntEnum):
    """Custom dynamically picklable IntEnum class.

    The Enum values for a device are created dynamically in toolkit based on
    the node informations. Since they are not predefined but rather created
    dynamically, the are not picklable. This custom child class of IntEnum
    overwrites the reduce function that returns all information required to
    recreate the Enum class in `NodeEnumMeta`.

    For more information on the reduce functionality and how it is used within
    the pickle package see
    [pep307](https://peps.python.org/pep-0307/#extended-reduce-api).
    """

    # Required for typing
    def __init__(self, *args, **kwargs):
        ...

    # Required for typing
    def __call__(self, *args, **kwargs):  # noqa: D102
        ...

    def __reduce_ex__(self, _):
        return NodeEnumMeta, (
            self._value_,
            self.__class__.__name__,
            {key: int(value) for key, value in self.__class__._member_map_.items()},
            self.__class__.__module__,
        )


class NodeInfo:
    """Class that holds the additional information for a single node.

    LabOne provides for each leaf node a dictionary of so called node info
    (In addition toolkit or the user may also add some information for selected
    nodes). This class wraps around these information and exposes them to a user
    in a friendly way.

    During the initialization is fetches the relevant information from the
    root of the nodetree and stores them internally.

    Args:
        node: A node the information belong to.
    """

    def __init__(self, node: "Node"):
        self._info = {}
        self._is_wildcard = False
        self._is_partial = False
        if any(wildcard in "".join(node.raw_tree) for wildcard in ["*", "?", "["]):
            self._info = {"Node": node.root.node_to_raw_path(node)}
            self._is_wildcard = True
        else:
            try:
                self._info = next(iter(node.root.get_node_info(node).values()))
                self._info["Node"] = self._info.get(
                    "Node", node.root.node_to_raw_path(node)
                ).lower()
            except KeyError as error:
                self._info = {"Node": error.args[0]}
                if self._check_partial(node):
                    self._is_partial = True
                if self._check_dynamic(node):
                    self._info.update(
                        json.loads(
                            node.root.connection.listNodesJSON(error.args[0])
                        ).get(error.args[0], {})
                    )

    def __dir__(self):
        dir_info = []
        for var, value in vars(self.__class__).items():
            if isinstance(value, property) and not var.startswith("_"):
                dir_info.append(var)
        return dir_info

    def __repr__(self) -> str:
        node_type = "partial" if self.is_partial else "leaf"
        node_type = "wildcard" if self.contains_wildcards else node_type
        return f'NodeInfo("{self.path}",{node_type}-node)'

    def __str__(self) -> str:
        string = self.path
        if self.is_partial:
            return string + "\nPartial node"
        if self._is_wildcard:
            return string + "\nContains wildcards"
        if "Description" in self._info:
            string += "\n" + self._info["Description"]
        for key, value in self._info.items():
            if key == "Options":
                string += f"\n{key}:"
                for option, description in value.items():
                    string += f"\n    {option}: {description}"
            elif key not in ["Description", "Node", "SetParser", "GetParser"]:
                string += f"\n{key}: {value}"
        return string

    def __getitem__(self, item: str) -> t.Union[str, t.Dict[str, str]]:
        return self._info[item]

    def __contains__(self, k):
        return k in self._info

    def __hash__(self):
        return hash(self.path + "NodeInfo")

    def __eq__(self, other):
        return self._info == other._info

    T = t.TypeVar("T")

    def set_parser(self, value: T) -> T:
        """Parse the set value."""
        try:
            _parser = self._info["SetParser"]
            if isinstance(_parser, list):
                for parser in _parser:
                    value = parser(value)
                return value
            return _parser(value)
        except KeyError:
            return value

    def get_parser(self, value: T) -> T:
        """Parse the get value."""
        try:
            _parser = self._info["GetParser"]
            if isinstance(_parser, list):
                for parser in _parser:
                    value = parser(value)
                return value
            return _parser(value)
        except KeyError:
            return value

    @staticmethod
    def _check_partial(node: "Node") -> bool:
        """Flag if the node is a partial node."""
        for child_node, _ in node.root:
            if node.is_child_node(child_node) and node != child_node:
                return True
        return False

    @staticmethod
    def _check_dynamic(node: "Node") -> bool:
        try:
            return node.raw_tree[-2] == "waves"
        except IndexError:
            return False

    @property
    def is_partial(self) -> bool:
        """Flag if the node is a partial node (non-leaf node)."""
        return self._is_partial

    @property
    def contains_wildcards(self) -> bool:
        """Flag if the node contains wildcards."""
        return self._is_wildcard

    @property
    def readable(self) -> t.Optional[bool]:
        """Flag if the node is readable.

        Returns None if the node does not provide the information (e.g. wildcard
        or partial nodes.)
        """
        try:
            return "Read" in self._info["Properties"]
        except KeyError:
            return None

    @property
    def writable(self) -> t.Optional[bool]:
        """Flag if the node is writable.

        Returns None if the node does not provide the information (e.g. wildcard
        or partial nodes.)
        """
        try:
            return "Write" in self._info["Properties"]
        except KeyError:
            return None

    @property
    def is_setting(self) -> t.Optional[bool]:
        """Flag if the node is setting node.

        Returns None if the node does not provide the information (e.g. wildcard
        or partial nodes.)
        """
        try:
            return "Setting" in self._info["Properties"]
        except KeyError:
            return None

    @property
    def is_vector(self) -> t.Optional[bool]:
        """Flag if the value of the node a vector.

        Returns None if the node does not provide the information (e.g. wildcard
        or partial nodes.)
        """
        try:
            return "Vector" in self._info["Type"]
        except KeyError:
            return None

    @property
    def path(self) -> str:
        """Path (LabOne representation) of the node."""
        return self._info["Node"]

    @property
    def description(self) -> str:
        """Description of the node."""
        return self._info["Description"]

    @property
    def type(self) -> str:
        """Type of the node."""
        return self._info["Type"]

    @property
    def unit(self) -> str:
        """Unit of the node."""
        return self._info["Unit"]

    _option_info = namedtuple("_option_info", ["enum", "description"])

    @lazy_property
    def options(self) -> t.Dict[int, _option_info]:
        """Options of the node."""
        option_map = {}
        for key, value in self._info.get("Options", {}).items():
            node_options = re.findall(r'("(.+?)"[,:]+)? ?(.*)', value)
            option_map[int(key)] = self._option_info(
                node_options[0][1], node_options[0][2]
            )
        return option_map

    @lazy_property
    def enum(self) -> t.Optional[NodeEnum]:
        """Enum of the node options."""
        try:
            options_reversed = {value.enum: key for key, value in self.options.items()}
            return (
                NodeEnum(self.path, options_reversed, module=__name__)
                if options_reversed
                else None
            )
        except ValueError:
            # Nameless options do not have a enum.
            return None


class Node:
    """Lazy node of a ``Nodetree``.

    The node is implemented in a lazy way. Meaning unless operations are not
    performed on the node, no checks whether the node is valid or not are
    performed.

    The child nodes of each node can be accessed in the same way than on the
    Nodetree, either by attribute or by item.

    The core functionality of each node is the overloaded call operator.
    Making a call gets the value(s) for that node. Passing a value to the call
    operator will set that value to the node on the device. Calling a node that
    is not a leaf (wildcard or partial node) will return/set the value on every
    node that matches it (the return type will be a dictionary).

    Warning:
        Setting a value to a non-leaf node will try to set the value of all
        nodes that matches that node. It should therefor be used with great care
        to avoid unintentional changes.

    >>> nodetree.demods[0].freq()
    1000
    >>> nodetree.demods[0].freq(2000)
    >>> nodetree.demods[0].freq()
    2000

    >>> nodetree.demods["*"].freq(3000)
    >>> nodetree.demods["*"].freq()
    {
        '/dev1234/demods/0/freq': 3000
        '/dev1234/demods/1/freq': 3000
        '/dev1234/demods/2/freq': 3000
        '/dev1234/demods/3/freq': 3000
    }

    .. versionchanged:: 0.3.5

        Call operator returns :class:`WildcardResult` when wildcards are used in
        getting values.

    The call operator supports the following flags:

    * deep: Flag if the set operation should be blocking until the data
        has arrived at the device, respectively if the get operation should
        return the value from the device or the cached value on the data
        server (if there is any). If this flag is set the operation can
        take significantly longer. (default = False)

        For a deep get operation the timestamp from the device is returned
        in addition to the value (The timestamp can be None, e.g. deep gets
        on LabOne modules).

        >>> nodetree.demods[0].freq(deep=True)
        (343283971893, 2000)

        For a deep set the call operator will return the value acknowledged
        by the device. e.g. important for floating point values with a
        limited resolution.

        >>> nodetree.demods[0].freq(29999,99999, deep=True)
        3000

    Warning:
        The deep flag does not work for wildcard nodes or non leaf nodes since
        they represent multiple nodes that are set in a transactional set which
        does not report the acknowledged values.

    * enum: Flag if enumerated values should return the enum value as
        string. (default = True)

    * parse: Flag if the SetParser/GetParser from the Node, if present,
        should be applied or not (default = True).

        The parsers are hard coded lambda functions provided not directly by
        LabOne but need to be set manually (e.g. toolkit adds these for a
        selected set of nodes). The Lambda function gets called right
        before/after the API call to LabOne. Usually they are used to add
        additional limitations and improve error reporting but in theory they
        can be used for anything.

        To add a parser to a node use the ``NodeTree.update_node`` function.

        >>> nodetree.update_node(
                "/dev1234/demods/0/freq",
                {
                    "GetParser": lambda v: print(f"got {v} from LabOne") return v,
                    "SetParser": lambda v: print(f"set {v} to LabOne") return v,
                },
            )

    In addition to the call operator the following magic methods are
    implemented:

    * __contains__
        >>> "freq" in nodetree.demods[0]
        True
    * __iter__
        >>> for node, info in nodetree.demods["*"].freq
    * __eq__
        >>> nodetree.demods[0].freq == nodetree.demods["*/freq"]
    * __len__ (only implemented for list like nodes)
        >>> len(nodetree.demods)
        4
    * __bool__ test if the node is a existing node
        >>> if nodetree.demods[0].freq:
            ...
    * __hash__ (e.g. necessary to be able to use nodes as key in dictionaries)

    Args:
        root: Root of the nodetree.
        tree: Tree (node path as tuple) of the current node.
    """

    def __init__(self, root: "NodeTree", tree: tuple):
        self._root = root
        self._tree = tree

    def __getattr__(self, name) -> "Node":
        return Node(self._root, self._tree + (name,))

    def __getitem__(self, name) -> "Node":
        name = str(name).lower()
        if "/" in name:
            name_list = name.split("/")
            if name_list[0]:
                return Node(self._root, self._tree + (*name_list,))
            return Node(self._root, self._tree + (*name_list[1:],))
        return Node(self._root, self._tree + (name,))

    def __contains__(self, k):
        return k in self._next_layer

    def __iter__(self):
        for child_node, info in self._root:
            if self.is_child_node(child_node):
                yield child_node, info

    def __repr__(self):
        return self.node_info.path

    def __dir__(self):
        dir_info = list(self._next_layer)
        for var, value in vars(self.__class__).items():
            if isinstance(value, property) and not var.startswith("_"):
                dir_info.append(var)
        return dir_info

    def __eq__(self, other):
        # buildin keywords are escaped with a tailing underscore
        # (https://pep8.org/#descriptive-naming-styles)
        own_node_list = tuple(node.rstrip("_") for node in self._tree)
        other_node_list = tuple(node.rstrip("_") for node in other.raw_tree)
        return own_node_list == other_node_list and self._root is other._root

    def __hash__(self):
        own_node_list = tuple(node.rstrip("_") for node in self._tree)
        if not own_node_list:
            own_node_list = "Node"
        return hash((own_node_list, repr(self._root)))

    def __bool__(self):
        return self.is_valid()

    def __len__(self):
        if not self._is_list():
            raise TypeError(f"Node {self.node_info.path} is not a list")
        return len(self._next_layer)

    def __call__(
        self, value: t.Any = None, *, deep=False, enum=True, parse=True, **kwargs
    ) -> t.Any:
        """Call operator that either gets (empty) or gets the value of a node.

        Args:
            value: Optional value that should be set to the node. If not
                specified the operator will return the value of the node
                instead.
            deep: Flag if the operation should block until the device has
                acknowledged the operation. The operation returns the value
                acknowledged by the device. This takes significantly longer
                than a normal operation and should be used carefully.
            enum: Flag if enumerated values should return the enum value as
                string or return the raw number.
            parse: Flag if the GetParser or SetParser, if present, should be
                applied or not.

        Returns:
            Value of the node for a get operation. If the deep flag is set the
            acknowledged value from the device is returned (applies also for
            the set operation).

            .. versionchanged:: 0.3.5

                Returns :class:`WildcardResult` when wildcards are used in
                getting values.

        Raises:
            AttributeError: If the connection does not support the necessary
                function to get/set the value.
            RuntimeError: If self.node_info.type if one of the following:
                [ZIPWAWave, ZITriggerSample, ZICntSample, ZIImpedanceSample,
                ZIScopeWave, ZIAuxInSample]. The reason is that these nodes can
                only be polled.
            TypeError: if the deep command is not available for this node
                (e.g. sample nodes)
            KeyError: If the node does not resolve to at least one valid leaf
                node.
        """
        if value is None:
            return self._get(deep=deep, enum=enum, parse=parse, **kwargs)
        return self._set(value, deep=deep, enum=enum, parse=parse, **kwargs)

    @lazy_property
    def _next_layer(self) -> t.Set[str]:
        """A set of direct child nodes."""
        next_layer = set()
        for node, _ in self:
            next_layer.add(node.raw_tree[len(self._tree)])
        return next_layer

    def _is_list(self) -> bool:
        """Checks if the node is a list type."""
        return len(self._next_layer) > 0 and next(iter(self._next_layer)).isdecimal()

    def _resolve_wildcards(self) -> t.List[str]:
        """Resolves potential wildcards.

        Also will resolve partial nodes to its leaf nodes.

        Returns:
            List of matched nodes in the raw path format
        """
        return fnmatch.filter(
            self._root.raw_dict.keys(), self._root.node_to_raw_path(self) + "*"
        )

    def _parse_get_value(
        self, value: t.Any, enum: bool = True, parse: bool = True
    ) -> t.Any:
        """Parse the raw value from the data server.

        Args:
            value: Raw value from the data server.
            enum: Flag if enumerated values should return the enum value as a
                string or as a raw number.
            parse: Flag if the GetParser, if present, should be applied or not.

        Returns:
            Parsed value
        """
        if enum and isinstance(value, int):
            try:
                value = self.node_info.enum(value) if self.node_info.enum else value
            except ValueError:
                # If the value is not in the enum LabOne does weird stuff but
                # we should not raise an exception ...
                pass
        if parse:
            value = self.node_info.get_parser(value)
        return value

    def _get(
        self, deep: bool = False, enum: bool = True, parse: bool = True, **kwargs
    ) -> t.Any:
        """Get the value from the node.

        The kwargs will be forwarded to the mapped zhinst.core function call.

        Args:
            deep: Flag if the get operation should return the cached value
                from the Data Server or get the value from the device, which is
                significantly slower.
            enum: Flag if enumerated values should return the enum value as
                string or return the raw number.
            parse: Flag if the GetParser, if present, should be applied or not.

        Return:
            value(s) from the device. If multiple values matches the the node a
            dictionary of the childnodes and their value is returned. If the
            ``deep`` flag is set the value is a pair (timestamp, value) instead.

        Raises:
            AttributeError: if the connection does not support the necessary
                function the get the value.
            RuntimeError: If self.node_info.type if one of the following:
                [ZIPWAWave, ZITriggerSample, ZICntSample, ZIImpedanceSample,
                ZIScopeWave, ZIAuxInSample]. The reason is that these nodes can
                only be polled.
            TypeError: if the deep command is not available for this node
                (e.g. sample nodes) or connection object.
            KeyError: If the node does not resolve to at least one valid leaf
                node.
        """
        readable = self.node_info.readable
        if readable:
            timestamp = None
            if deep:
                timestamp, value = self._get_deep(**kwargs)
            else:
                value = self._get_cached(**kwargs)
            value = self._parse_get_value(value, enum=enum, parse=parse)
            return (timestamp, value) if deep else value
        if readable is None and (
            self.node_info.contains_wildcards or self.node_info.is_partial
        ):
            return self._get_wildcard(deep=deep, enum=enum, parse=parse, **kwargs)
        if readable is False:
            raise AttributeError(f"{self.node_info.path} is not readable.")
        raise KeyError(self.node_info.path)

    @staticmethod
    def _parse_get_entry(raw_value: t.Dict[t.Union[str, int], t.Any]):
        """Parser for the get function of zhinst.core.

        The get function in zhinst.core support multiple values and returns the
        results as a OrderdDict. This functions parses the value entry of that
        dictionary into a (timestamp, value) pair.

        Args:
            raw_value: OrderdDict from the zhinst.core get command.

        Returns:
            (timestamp, value) pair.
        """
        value = None
        timestamp = None
        try:
            value = raw_value["value"][0]
            timestamp = raw_value["timestamp"][0]
        except TypeError:
            # ZIVectorData have a different structure
            value = raw_value[0]
            if isinstance(value, dict):
                timestamp = value["timestamp"]
                value = value["vector"]
        except IndexError:
            # HF2 has not timestamp
            value = raw_value[0]
        except KeyError:
            # HF2 returns sample nodes as well but we don`t parse them
            value = raw_value
        return (timestamp, value)

    def _get_wildcard(
        self, deep=True, enum=True, parse=True, **kwargs
    ) -> t.Union["WildcardResult", t.Dict[str, t.Any]]:
        """Execute a wildcard get.

        The get is performed as a deep get (for all devices except HF2)
        regardless of the ``deep`` flag. If the ``deep`` flag is not set the
        timestamp is removed to ensure concistency.

        The kwargs will be forwarded to the maped zhinst.core function call.

        Args:
            deep: Flag if the get operation should return the cached value
                from the Data Server or get the value from the device, which is
                significantly slower. The wildcard get is always performed as
                deep get but the timestamp is only returned if the ``deep```
                flag is set.
            enum: Flag if enumerated values should return the enum value as
                string or return the raw number.
            parse: Flag if the GetParser, if present, should be applied or not.

        Returns:
            ``WildcardResult`` if deep is `True`. Else a dictionary.
            Dictionary or a Mapping with the values of all subnodes.

        Raises:
            KeyError: If the node does not resolve to at least one valid leaf
                node.
        """
        # modules don`t have settingsonly argument ... this will be caught in
        # a try catch block to avoid unnecessary comparisons
        kwargs.setdefault("settingsonly", False)
        kwargs.setdefault("flat", True)
        try:
            result_raw = self._root.connection.get(self.node_info.path, **kwargs)
        except TypeError:
            del kwargs["settingsonly"]
            try:
                result_raw = self._root.connection.get(self.node_info.path, **kwargs)
            except (RuntimeError, TypeError):
                # resolve wildecard and get the value of the resulting leaf nodes
                nodes_raw = self._resolve_wildcards()
                result_raw = self._root.connection.get(",".join(nodes_raw), **kwargs)
        except RuntimeError:
            # resolve wildecard and get the value of the resulting leaf nodes
            nodes_raw = self._resolve_wildcards()
            result_raw = self._root.connection.get(",".join(nodes_raw), **kwargs)
        if not result_raw:
            raise KeyError(self.node_info.path)
        if not kwargs["flat"]:
            return result_raw
        result = {}
        for sub_node_raw, node_value in result_raw.items():
            sub_node = self._root.raw_path_to_node(sub_node_raw)
            timestamp, value = self._parse_get_entry(node_value)
            value = sub_node._parse_get_value(value, enum=enum, parse=parse)
            # although the operation is a deep get we hide the timestamp
            # to ensure consistency
            result[sub_node_raw] = (timestamp, value) if deep else value
        return WildcardResult(result)

    def _get_deep(self, **kwargs) -> t.Tuple[int, t.Any]:
        """Get the node value from the device.

        The kwargs will be forwarded to the maped zhinst.core function call.

        Note: The HF2 does not support the timestamp option and will therfore
        return None for the timestamp.

        Returns:
            (timestamp, value) from the device.

        Raises:
            TypeError: if the deep command is not available for this node
                (e.g. sample nodes)
        """
        kwargs.setdefault("settingsonly", False)
        # Flat must be set to True (if customers want the flat option they need
        # to call zhinst.core directly)
        kwargs["flat"] = True
        raw_dict = self._root.connection.get(self.node_info.path, **kwargs)
        if not raw_dict or len(raw_dict) == 0:
            raise TypeError(
                "keyword 'deep' is not available for this node. "
                "(e.g. node is a sample node)"
            )
        raw_value = next(iter(raw_dict.values()))
        return self._parse_get_entry(raw_value)

    def _get_cached(self, **kwargs) -> t.Any:
        """Get the cached node value from the data server.

        The kwargs will be forwarded to the maped zhinst.core function call.

        Returns:
            Cached node value from the data server.

        Raises:
            AttributeError: if the connection does not support the necessary
                function the get the value.
            RuntimeError: If self.node_info.type if one of the following:
                [ZIPWAWave, ZITriggerSample, ZICntSample, ZIImpedanceSample,
                ZIScopeWave, ZIAuxInSample]. The reason is that these nodes can
                only be polled.
        """
        if "Integer" in self.node_info.type:
            return self._root.connection.getInt(self.node_info.path)
        if self.node_info.type == "Double":
            return self._root.connection.getDouble(self.node_info.path)
        if self.node_info.type == "String":
            return self._root.connection.getString(self.node_info.path)
        if self.node_info.type == "ZIVectorData":
            _, value = self._get_deep(**kwargs)
            return value
        if self.node_info.type == "Complex Double":
            return self._root.connection.getComplex(self.node_info.path, **kwargs)
        if self.node_info.type == "ZIDemodSample":
            return self._root.connection.getSample(self.node_info.path, **kwargs)
        if self.node_info.type == "ZIDIOSample":
            return self._root.connection.getDIO(self.node_info.path, **kwargs)
        if self.node_info.type == "ZIAdvisorWave":
            raw_value = self._root.connection.get(self.node_info.path, flat=True)
            return next(iter(raw_value.values()))[-1]
        raise RuntimeError(
            f"{self.node_info.path} has type {self.node_info.type} and can "
            "only be polled."
        )

    def _set(
        self, value: t.Any, deep=False, enum=True, parse=True, **kwargs
    ) -> t.Optional[t.Any]:
        """Set the value to the node.

        The kwargs will be forwarded to the maped zhinst.core function call.

        Args:
            value: value
            deep: Flag if the set operation should be blocking until the data
                has arrived at the device. (default=False)
            enum: Flag if enumerated values should accept the enum value as
                string. (default=True)
            parse: Flag if the SetParser, if present, should be applied or not.
                (default=True)

        Returns:
            Acknowledged value on the device if ``deep`` flag is set (does not
            work for wildcard or non leafs nodes since they are bundled in a
            transaction).

        Raises:
            AttributeError: If the connection does not support the necessary
                function the set the value.
            KeyError: if the wildcard does not resolve to a valid node
            RuntimeError: if deep set is not possible
            TypeError: Connection does not support deep set
        """
        writable = self.node_info.writable
        if writable:
            if parse:
                value = self.node_info.set_parser(value)
            if self._root.transaction.in_progress():
                self._root.transaction.add(self, value)
            elif deep:
                return self._parse_get_value(
                    self._set_deep(value, **kwargs), enum=enum, parse=parse
                )
            else:
                try:
                    self._root.connection.set(self.node_info.path, value, **kwargs)
                except RuntimeError:
                    # Some vector nodes do not support support set command.
                    if self.node_info.type == "ZIVectorData":
                        self._root.connection.setVector(
                            self.node_info.path, value, **kwargs
                        )
                    else:
                        raise
            return None
        if writable is None and (
            self.node_info.contains_wildcards or self.node_info.is_partial
        ):
            self._set_wildcard(value, parse=parse, **kwargs)
            return None
        if writable is False:
            raise AttributeError(f"{self.node_info.path} is read-only.")
        raise KeyError(self.node_info.path)

    def _set_wildcard(self, value: t.Any, parse: bool = True, **kwargs) -> None:
        """Performs a transactional set on all nodes that match the wildcard.

        The kwargs will be forwarded to the mapped zhinst.core function call.

        Args:
            value: value
            parse: Flag if the SetParser, if present, should be applied or not.
                (default=True)

        Raises:
            KeyError: if the wildcard does not resolve to a valid node
        """
        nodes_raw = self._resolve_wildcards()
        if not nodes_raw:
            raise KeyError(self._root.node_to_raw_path(self))
        with create_or_append_set_transaction(self._root):
            for node_raw in nodes_raw:
                self._root.raw_path_to_node(node_raw)(value, parse=parse, **kwargs)

    def _set_deep(self, value: t.Any, **kwargs) -> t.Any:
        """Set the node value from device.

        The kwargs will be forwarded to the mapped zhinst.core function call.

        Args:
            value: value

        Returns:
            Acknowledged value on the device. Only if a deeps set operation is
            available for the connection object.

        Raises:
            RuntimeError: if deep set is not possible
            TypeError: Connection does not support deep set
        """
        try:
            if isinstance(value, numbers.Integral):
                return self._root.connection.syncSetInt(
                    self.node_info.path, value, **kwargs
                )
            if isinstance(value, numbers.Real):
                return self._root.connection.syncSetDouble(
                    self.node_info.path, value, **kwargs
                )
            if isinstance(value, str):
                return self._root.connection.syncSetString(
                    self.node_info.path, value, **kwargs
                )
        except TypeError as error:
            raise TypeError(
                "deep set is not supported for this connection."
                "(this likely cause because the connection is a module and a deep "
                "set does not make sense there.)"
            ) from error
        raise RuntimeError(
            f"Invalid type {type(value)} for deep set "
            "(only int,float and str are supported)"
        )

    def is_child_node(self, child_node: "Node") -> bool:
        """Checks if a node is child node of this node.

        Args:
            child_node: Potential child node

        Returns:
            Boolean if passed node is a child node
        """
        return fnmatch.fnmatchcase(
            "/".join(child_node.raw_tree), "/".join(self.raw_tree) + "*"
        )

    def wait_for_state_change(
        self,
        value: t.Union[int, str],
        *,
        invert: bool = False,
        timeout: float = 2,
        sleep_time: float = 0.005,
    ) -> None:
        """Waits until the node has the expected state/value.

        Warning:
            Only supports integer nodes. (The value can either be the value or
            its corresponding enum value as string)

        Args:
            value: Expected value of the node.
            invert: Instead of waiting for the value, the function will wait for
                any value except the passed value instead. (default = False)

                Useful when waiting for value to change from existing one.
            timeout: Maximum wait time in seconds. (default = 2)
            sleep_time: Sleep interval in seconds. (default = 0.005)

        Raises:
            TimeoutError: Timeout exceeded.
        """
        if self.node_info.contains_wildcards:
            start_time = time.time()
            nodes_raw = self._resolve_wildcards()
            if not nodes_raw:
                raise KeyError(self.node_info.path)
            for node_raw in nodes_raw:
                node = self._root.raw_path_to_node(node_raw)
                node.wait_for_state_change(
                    value,
                    invert=invert,
                    timeout=max(0, timeout - (time.time() - start_time)),
                    sleep_time=sleep_time,
                )
        else:
            value = self._parse_get_value(value)
            start_time = time.time()
            while (
                start_time + timeout >= time.time() and (self._get() == value) == invert
            ):
                time.sleep(sleep_time)
            curr_value = self._get()
            if (curr_value == value) is invert:
                if invert:
                    raise TimeoutError(
                        f"{self.node_info.path} did not change from the expected"
                        f" value {value} within {timeout}s."
                    )
                raise TimeoutError(
                    f"{self.node_info.path} did not change to the expected value"
                    f" within {timeout}s. {value} != {curr_value}"
                )

    def subscribe(self) -> None:
        """Subscribe to this node (its child lead nodes).

        To get data from data from the subscribed nodes use the poll command
        (provided by the Connection).

        In order to avoid fetching old data that is still in the buffer execute
        a sync command before subscribing to new data streams.
        """
        try:
            self._root.connection.subscribe(self.node_info.path)
        except RuntimeError as error:
            nodes_raw = self._resolve_wildcards()
            if not nodes_raw:
                raise KeyError(self.node_info.path) from error
            for node_raw in nodes_raw:
                self._root.connection.subscribe(node_raw)

    def unsubscribe(self) -> None:
        """Unsubscribe this node (its child lead nodes).

        Use this command after recording to avoid buffer overflows that may
        increase the latency of subsequent commands.
        """
        try:
            self._root.connection.unsubscribe(self.node_info.path)
        except RuntimeError as error:
            nodes_raw = self._resolve_wildcards()
            if not nodes_raw:
                raise KeyError(self.node_info.path) from error
            for node_raw in nodes_raw:
                self._root.connection.unsubscribe(node_raw)

    def get_as_event(self) -> None:
        """Trigger an event for that node (its child lead nodes).

        The node data is returned by a subsequent poll command.
        """
        try:
            self._root.connection.getAsEvent(self.node_info.path)
        except RuntimeError as error:
            nodes_raw = self._resolve_wildcards()
            if not nodes_raw:
                raise KeyError(self.node_info.path) from error
            for node_raw in nodes_raw:
                self._root.connection.getAsEvent(node_raw)

    def child_nodes(
        self,
        *,
        recursive: bool = False,
        leavesonly: bool = False,
        settingsonly: bool = False,
        streamingonly: bool = False,
        subscribedonly: bool = False,
        basechannelonly: bool = False,
        excludestreaming: bool = False,
        excludevectors: bool = False,
        full_wildcard: bool = False,
    ) -> t.Generator["Node", None, None]:
        """Generator for all child nodes that matches the filters.

        If the nodes does not contain any child nodes the generator will only
        contain the node itself (if it matches the filters).

        Warning:
            The Data Server supports only the asterisk wildcard. For all
            other wildcards the matching child nodes can be generated manually
            within the nodetree.

            Enabling additional flags how ever require that each node that
            matches the wildcard needs to be checked in a separate request by
            the Data Server which can cause a severe latency. Therefore one
            needs to enable the `full_wildcard` flag in order to support the
            manual generation of the matching child nodes.

        Examples:
            >>> child_nodes = nodetree.demods[0].child_nodes()
            >>> next(child_nodes)
            /dev1234/demods/0/freq

        Args:
            recursive: Returns the nodes recursively (default: False)
            leavesonly: Returns only nodes that are leaves, which means
                they are at the outermost level of the tree (default: False).
            settingsonly: Returns only nodes which are marked as setting
                (default: False).
            streamingonly: Returns only streaming nodes (default: False).
            subscribedonly: Returns only subscribed nodes
                (default: False).
            basechannelonly: Return only one instance of a node in case
                of multiple channels (default: False).
            excludestreaming: Exclude streaming nodes (default: False).
            excludevectors: Exclude vector nodes (default: False).
            full_wildcard: Enables full wildcard support. Per default
                only the asterisk wildcard is supported. (Automatically sets
                recursive and leavesonly) (default = False)

        Returns:
            Generator of all child nodes that match the filters
        """
        raw_path = self._root.node_to_raw_path(self)
        try:
            raw_result = self._root.connection.listNodes(
                raw_path,
                recursive=recursive,
                leavesonly=leavesonly,
                settingsonly=settingsonly,
                streamingonly=streamingonly,
                subscribedonly=subscribedonly,
                basechannelonly=basechannelonly,
                excludestreaming=excludestreaming,
                excludevectors=excludevectors,
            )
            for node_raw in raw_result:
                yield self._root.raw_path_to_node(node_raw)
        except RuntimeError as error:
            if "Path format invalid" not in error.args[0]:
                raise
            if not full_wildcard:
                raise RuntimeError(
                    "The node contains wildcards that the DataServer can not resolve. "
                    "Use the `full_wildcard` flag to search for child nodes manually."
                ) from error
            nodes_raw = self._resolve_wildcards()
            for node_raw in nodes_raw:
                node = self._root.raw_path_to_node(node_raw)
                if not (
                    (settingsonly and not node.node_info.is_setting)
                    or (excludevectors and node.node_info.is_vector)
                    or (
                        basechannelonly
                        and any(number != 0 for number in re.findall(r"\d+", node_raw))
                    )
                    or (
                        (excludestreaming or subscribedonly or streamingonly)
                        and not self._root.connection.listNodes(
                            node_raw[:-1] + "*",  # TODO remove once listNodes is fixed
                            recursive=recursive,
                            leavesonly=leavesonly,
                            settingsonly=settingsonly,
                            streamingonly=streamingonly,
                            subscribedonly=subscribedonly,
                            basechannelonly=basechannelonly,
                            excludestreaming=excludestreaming,
                            excludevectors=excludevectors,
                        )
                    )
                ):
                    yield node

    @lru_cache()
    def is_valid(self) -> bool:
        """Check if the node is a valid node.

        Valid node means it resolves to at least one existing node in the node
        tree. Meaning not only leaf nodes are valid nodes but also partial nodes
        and nodes containing wildcards.

        Returns:
            Flag if the node is a valid node
        """
        keys = self._resolve_wildcards()
        return len(keys) > 0

    @lazy_property
    def node_info(self) -> NodeInfo:
        """Additional information about the node."""
        return NodeInfo(self)

    @property
    def raw_tree(self) -> t.Tuple[str, ...]:
        """Internal representation of the node."""
        return self._tree

    @property
    def root(self) -> "NodeTree":
        """Node tree to which this node belongs to."""
        return self._root


class NodeList(Sequence, Node):
    """List of nodelike objects.

    List of preinitialized classes that inherit from the ``Node`` class would not
    support wildcards since they would be of type list.

    This class holds the preinitialized objects. But if a the passed item is not
    an integer it returns a Node instead.

    Warning:
        Since in case of a passed wildcard symbol the return value is a node,
        the additional functionality that the nodelike object may provide
        (e.g. helper functions) are no longer accessible.

    Args:
        elements: Preinitialized child elements
        root: Root of the nodetree
        tree: Node tree (node path as tuple) of the current node
    """

    def __init__(self, elements: t.Sequence[t.Any], root: "NodeTree", tree: tuple):
        Sequence.__init__(self)
        Node.__init__(self, root, tree)
        self._elements: t.Sequence[t.Any] = elements

    @t.overload
    def __getitem__(self, idx: t.Union[int, str]) -> t.Union[t.Any, Node]:
        ...

    @t.overload
    def __getitem__(self, s: slice) -> t.Sequence[t.Union[t.Any, Node]]:
        ...

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._elements[item]
        return Node(self._root, self._tree + (str(item),))

    def __len__(self):
        return len(self._elements)

    def __eq__(self, other):
        return Node.__eq__(self, other)

    def __hash__(self):
        return Node.__hash__(self)


class WildcardResult(Mapping):
    """Mapping of results when a :class:`Node` is called with wildcards.

    Args:
        result: A dictionary of node/value pairs.

    Example:
        >>> result = device.demods["*"].enable()
        >>> print(result)
        {
            '/dev1234/demods/0/enable': 0,
            '/dev1234/demods/1/enable': 1,
        }
        >>> result[device.demods[0].enable]
        0
        >>> result["/dev1234/demods/0/enable"]
        0

    .. versionadded:: 0.3.5
    """

    def __init__(self, result: t.Dict[str, t.Any]):
        self._result = result

    def __repr__(self):
        return repr(self._result)

    def __getitem__(self, key: t.Union[str, Node]):
        return self._result[str(key)]

    def __iter__(self):
        return iter(self._result)

    def __len__(self):
        return len(self._result)

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Convert the WildcardResult to a dictionary.

        After conversion, :class:`Node` objects cannot be used to get items.
        """
        return self._result
